from django.db import models
from model_utils import FieldTracker


class Agent(models.Model):
    name = models.CharField(max_length=255)
    is_disabled = models.BooleanField(default=False)
    email = models.EmailField(blank=True, null=True)
    initials = models.CharField(max_length=10, blank=True, null=True)
    firstname = models.CharField(max_length=255, blank=True, null=True)
    surname = models.CharField(max_length=255, blank=True, null=True)
    colour = models.CharField(max_length=255, blank=True, null=True)
    teams = models.ManyToManyField('Team', related_name='agents', blank=True)

    class Meta:
        verbose_name_plural = "Agents"

    def __str__(self):
        return f"{self.firstname} {self.surname}"


class AgentTracker(Agent):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'halo_agent'



from django.db import models
from model_utils import FieldTracker


class ChargeRate(models.Model):
    # Currently charge rates are only used for creating time entries.
    # We don't need the rate data or anything, but the user needs to
    # select one to create a time sheet event.
    name = models.CharField(max_length=255)

    def __str__(self):
        return str(self.name)


class ChargeRateTracker(ChargeRate):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'halo_chargerate'



# Generated by Django 4.2.16 on 2025-02-26 14:08

from django.db import migrations, models
import django.db.models.deletion


def clear_outcome(apps, _):
    Action = apps.get_model('halo', 'Action')
    Action.objects.update(outcome=None)

class Migration(migrations.Migration):

    dependencies = [
        ('halo', '0023_action_non_billable_time_action_outcome_and_more'),
    ]

    operations = [
        migrations.RunPython(clear_outcome, reverse_code=migrations.RunPython.noop),
        migrations.CreateModel(
            name='Outcome',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('outcome', models.CharField(max_length=255)),
                ('button_name', models.CharField(max_length=255)),
                ('label_long', models.CharField(max_length=255)),
                ('sequence', models.IntegerField()),
                ('hidden', models.BooleanField(default=False)),
                ('icon', models.CharField(max_length=255)),
            ],
        ),
        migrations.CreateModel(
            name='OutcomeTracker',
            fields=[
            ],
            options={
                'db_table': 'halo_outcome',
                'proxy': True,
                'indexes': [],
                'constraints': [],
            },
            bases=('halo.outcome',),
        ),
        migrations.AlterField(
            model_name='action',
            name='outcome',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='halo.outcome'),
        ),
    ]



# This file is used to import all the models for the records module
#  from a single location.

from djpsa.halo.records.ticket.model import Ticket, TicketTracker
from djpsa.halo.records.priority.model import Priority, PriorityTracker
from djpsa.halo.records.status.model import Status, StatusTracker
from djpsa.halo.records.client.model import Client, ClientTracker
from djpsa.halo.records.agent.model import Agent, AgentTracker
from djpsa.halo.records.site.model import Site, SiteTracker
from djpsa.halo.records.sla.model import SLA, SLATracker
from djpsa.halo.records.appointment.model import Appointment, AppointmentTracker
from djpsa.halo.records.halouser.model import HaloUser, HaloUserTracker
from djpsa.halo.records.tickettype.model import TicketType, TicketTypeTracker
from djpsa.halo.records.action.model import Action, ActionTracker
from djpsa.halo.records.team.model import Team, TeamTracker
from djpsa.halo.records.budgettype.model import BudgetType, BudgetTypeTracker
from djpsa.halo.records.budgetdata.model import BudgetData, BudgetDataTracker
from djpsa.halo.records.outcome.model import Outcome, OutcomeTracker
from djpsa.halo.records.chargerate.model import ChargeRate, ChargeRateTracker



from djpsa.halo.api import HaloAPIClient


UNASSIGNED_AGENT_ID = 1  # The Agent with id 1 represents the concept of being
# unassigned. Apparently the designer of this API dropped out before the
# lesson on NULLs.


class AgentAPI(HaloAPIClient):
    endpoint = 'Agent'



import unittest
import requests
from unittest.mock import patch, MagicMock
from json import JSONDecodeError

from djpsa.api.client import APIClient
from djpsa.api import exceptions as exc


class TestAPIClient(unittest.TestCase):

    def setUp(self):
        self.client = APIClient()
        self.client.request_settings = {
            'max_attempts': 3,
            'timeout': 5
        }

    @patch('djpsa.api.client.APIClient.request')
    def test_fetch_resource_success(self, mock_request):
        mock_request.return_value = {'data': 'test'}
        response = \
            self.client.fetch_resource('http://example.com/api/resource')
        self.assertEqual(response, {'data': 'test'})

    @patch('djpsa.api.client.APIClient.request')
    def test_fetch_resource_api_error(self, mock_request):
        mock_request.side_effect = exc.APIError('API Error')
        with self.assertRaises(exc.APIError):
            self.client.fetch_resource('http://example.com/api/resource')

    @patch.object(
        APIClient, '_format_endpoint', return_value='http://example.com')
    @patch.object(APIClient, '_request')
    @patch.object(APIClient, '_get_headers', return_value={})
    @patch.object(
        APIClient, '_format_params', return_value={'param': 'value'})
    def test_request_success(self, _, _a, mock_request, mock_format_endpoint):
        client = APIClient()
        mock_response = MagicMock()
        mock_response.status_code = 200
        mock_response.headers = {'Content-Type': 'application/json'}
        mock_response.json.return_value = {'key': 'value'}
        mock_request.return_value = mock_response

        response = client.request('GET')

        self.assertEqual(response, {'key': 'value'})
        mock_format_endpoint.assert_called_once()
        mock_request.assert_called_once_with(
            'GET', 'http://example.com', headers={}, params={'param': 'value'})

    @patch.object(
        APIClient, '_format_endpoint', return_value='http://example.com')
    @patch.object(APIClient, '_request')
    @patch.object(APIClient, '_get_headers', return_value={})
    @patch.object(APIClient, '_format_params', return_value={'param': 'value'})
    def test_request_json_decode_error(
            self, _, _a, mock_request, mock_format_endpoint):
        client = APIClient()
        mock_response = MagicMock()
        mock_response.status_code = 200
        mock_response.headers = {'Content-Type': 'application/json'}
        mock_response.json.side_effect = \
            JSONDecodeError('Expecting value', 'doc', 0)
        mock_request.return_value = mock_response

        with self.assertRaises(exc.APIError):
            client.request('GET')

        mock_format_endpoint.assert_called_once()
        mock_request.assert_called_once_with(
            'GET', 'http://example.com', headers={}, params={'param': 'value'})

    @patch.object(
        APIClient, '_format_endpoint', return_value='http://example.com')
    @patch.object(APIClient, '_request')
    @patch.object(APIClient, '_get_headers', return_value={})
    @patch.object(APIClient, '_format_params', return_value={'param': 'value'})
    def test_request_request_exception(
            self, _, _a, mock_request, mock_format_endpoint):
        client = APIClient()
        mock_request.side_effect = requests.RequestException('Request failed')

        with self.assertRaises(exc.APIError):
            client.request('GET')

        mock_format_endpoint.assert_called_once()
        mock_request.assert_called_once_with(
            'GET', 'http://example.com', headers={}, params={'param': 'value'})

    @patch.object(
        APIClient, '_format_endpoint', return_value='http://example.com')
    @patch.object(APIClient, '_request')
    @patch.object(APIClient, '_get_headers', return_value={})
    @patch.object(APIClient, '_format_params', return_value={'param': 'value'})
    def test_request_unexpected_code(
            self, _, _a, mock_request, mock_format_endpoint):
        client = APIClient()
        mock_response = MagicMock()
        mock_response.status_code = 418
        mock_request.return_value = mock_response

        with self.assertRaises(exc.APIError):
            client.request('GET')

        mock_format_endpoint.assert_called_once()
        mock_request.assert_called_once_with(
            'GET', 'http://example.com', headers={}, params={'param': 'value'})



from django.db import models
from model_utils import FieldTracker


class HaloUser(models.Model):

    # This is their username, even though it comes through the API
    # as name.
    name = models.CharField(max_length=255, blank=True, null=True)

    first_name = models.CharField(max_length=255, blank=True, null=True)
    initials = models.CharField(max_length=10, blank=True, null=True)
    surname = models.CharField(max_length=255, blank=True, null=True)
    email = models.EmailField(null=True, blank=True, max_length=250)
    colour = models.CharField(max_length=50, null=True, blank=True)
    active = models.BooleanField(default=True)
    login = models.CharField(max_length=255, blank=True, null=True)
    use = models.CharField(max_length=255, blank=True, null=True)
    client = models.ForeignKey(
        'Client', blank=True, null=True, on_delete=models.CASCADE)
    agent = models.ForeignKey(
        'Agent', blank=True, null=True, on_delete=models.CASCADE)
    site = models.ForeignKey(
        'Site', blank=True, null=True, on_delete=models.SET_NULL)
    never_send_emails = models.BooleanField(default=False)
    phone_number = models.CharField(blank=True, null=True, max_length=250)
    mobile_number = models.CharField(blank=True, null=True, max_length=250)
    mobile_number_2 = models.CharField(blank=True, null=True, max_length=250)
    home_number = models.CharField(blank=True, null=True, max_length=250)
    tel_pref = models.IntegerField(blank=True, null=True)
    is_service_account = models.BooleanField(default=False)
    is_important_contact = models.BooleanField(default=False)
    is_important_contact_2 = models.BooleanField(default=False)

    class Meta:
        verbose_name_plural = "Halo Users"

    def __str__(self):
        return str(self.name)


class HaloUserTracker(HaloUser):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'halo_halouser'



from typing import Any, List

from djpsa.halo import models
from djpsa.halo.records import api
from djpsa.halo import sync


class HaloUserSynchronizer(sync.ResponseKeyMixin, sync.HaloSynchronizer):
    model_class = models.HaloUserTracker
    response_key = 'users'
    client_class = api.UserAPI

    related_meta = {
        'client_id': (models.Client, 'client'),
        'linked_agent_id': (models.Agent, 'agent'),
    }

    def __init__(self,
                 full: bool = False,
                 conditions: List = None,
                 *args: Any,
                 **kwargs: Any):
        super().__init__(full, conditions, *args, **kwargs)

        self.client.add_condition({
            'includeactive': True,
        })

    def _assign_field_data(self, instance, json_data):
        instance.id = json_data.get('id')
        instance.name = json_data.get('name')
        instance.first_name = json_data.get('firstname')
        instance.surname = json_data.get('surname')
        instance.initials = json_data.get('initials')
        instance.email = json_data.get('emailaddress')
        instance.colour = json_data.get('colour')
        instance.active = not json_data.get('inactive', True)
        instance.login = json_data.get('login')
        instance.use = json_data.get('use')
        instance.never_send_emails = json_data.get('neversendemails', False)
        instance.phone_number = json_data.get('phonenumber')
        instance.mobile_number = json_data.get('mobilenumber')
        instance.mobile_number_2 = json_data.get('mobilenumber2')
        instance.home_number = json_data.get('homenumber')
        instance.tel_pref = json_data.get('telpref')
        instance.is_service_account = json_data.get('isserviceaccount', False)
        instance.is_important_contact = \
            json_data.get('isimportantcontact', False)
        instance.is_important_contact_2 = \
            json_data.get('isimportantcontact2', False)

        if json_data.get('linked_agent_id') == 0:
            json_data['linked_agent_id'] = None

        self.set_relations(instance, json_data)



import unittest
from unittest.mock import patch, MagicMock
from djpsa.halo.api import HaloAPIClient


class TestHaloAPIClient(unittest.TestCase):

    @patch('djpsa.halo.api.requests.request')
    @patch('djpsa.halo.api.HaloAPITokenFetcher.get_token',
           return_value='test_token')
    def test_request_success(self, _, mock_request):
        mock_response = MagicMock()
        mock_response.status_code = 200
        mock_request.return_value = mock_response

        client = HaloAPIClient()
        response = client._request('GET', 'http://example.com')

        self.assertEqual(response, mock_response)
        mock_request.assert_called_once_with(
            'GET',
            'http://example.com',
            headers={'Authorization': 'Bearer test_token'},
            params=None,
            timeout=30.0
        )

    @patch('djpsa.halo.api.requests.request')
    @patch('djpsa.halo.api.HaloAPITokenFetcher.get_token')
    def test_request_token_refresh(
            self, mock_get_token, mock_request):
        mock_get_token.side_effect = ['expired_token', 'new_token']
        mock_response_401 = MagicMock()
        mock_response_401.status_code = 401
        mock_response_200 = MagicMock()
        mock_response_200.status_code = 200
        mock_request.side_effect = [mock_response_401, mock_response_200]

        client = HaloAPIClient()
        response = client._request('GET', 'http://example.com')

        self.assertEqual(response, mock_response_200)
        self.assertEqual(mock_get_token.call_count, 2)
        self.assertEqual(mock_request.call_count, 2)
        mock_request.assert_called_with(
            'GET',
            'http://example.com',
            headers={'Authorization': 'Bearer new_token'},
            params=None
        )



from typing import Any, List
from django.utils import timezone
from dateutil.parser import parse

from djpsa.halo import models
from djpsa.halo.records import api
from djpsa.halo import sync


class AppointmentSynchronizer(sync.CreateMixin,
                              sync.UpdateMixin,
                              sync.DeleteMixin,
                              sync.ResponseKeyMixin,
                              sync.HaloSynchronizer):
    response_key = 'appointments'
    model_class = models.AppointmentTracker
    client_class = api.AppointmentAPI

    related_meta = {
        'client_id': (models.Client, 'client'),
        'agent_id': (models.Agent, 'agent'),
        'site_id': (models.Site, 'site'),
        'user_id': (models.HaloUser, 'user'),
        'ticket_id': (models.Ticket, 'ticket'),
    }

    def __init__(
            self,
            full: bool = False,
            conditions: List = None,
            *args: Any,
            **kwargs: Any):
        super().__init__(full, conditions, *args, **kwargs)

        self.client.add_condition({
            'hidecompleted': True,
        })

    def _assign_field_data(self, instance, json_data):
        instance.id = json_data.get('id')
        instance.subject = json_data.get('subject')
        start_date = json_data.get('start_date')
        instance.start_date = timezone.make_aware(
            parse(start_date), timezone.utc
        ) if start_date else None

        end_date = json_data.get('end_date')
        instance.end_date = timezone.make_aware(
            parse(end_date), timezone.utc
        ) if end_date else None
        instance.appointment_type = json_data.get('appointment_type_name')
        instance.is_private = json_data.get('is_private')
        instance.is_task = json_data.get('is_task', False)
        instance.complete_status = json_data.get('complete_status')
        instance.colour = json_data.get('colour')
        instance.online_meeting_url = json_data.get('online_meeting_url')

        self.set_relations(instance, json_data)



from django.db import models
from model_utils import FieldTracker


class Team(models.Model):
    name = models.CharField(max_length=255)
    ticket_count = models.IntegerField(default=0)

    class Meta:
        verbose_name_plural = "Teams"

    def __str__(self):
        return str(self.name)


class TeamTracker(Team):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'halo_team'



import base64
import os

import requests

from djpsa.halo.api import HaloAPIClient

FILE_UMASK = 0o022


class AttachmentAPI(HaloAPIClient):
    endpoint = 'Attachment'

    def get(self, record_id):
        return self.request('GET', params={'ticket_id': record_id})

    def upload(self, ticket_id, filename, content_type, file_content):
        """
        Upload an attachment to a ticket.

        Args:
            ticket_id: The ticket ID to attach to
            filename: The filename
            content_type: The MIME type (e.g. 'text/plain', 'image/png')
            file_content: The file content as bytes

        Returns:
            The API response
        """
        base64_content = base64.b64encode(file_content).decode('utf-8')
        data_uri = f'data:{content_type};base64,{base64_content}'

        return self.request('POST', body=[{
            'ticket_id': ticket_id,
            'filename': filename,
            'data_base64': data_uri,
        }])

    def download_from_url(self, url, attachment_id, path):
        """
        Download attachment from CDN URL and save to disk.
        """
        response = requests.get(url)
        response.raise_for_status()

        saved_filename = str(attachment_id)
        file_path = os.path.join(path, saved_filename)

        previous_umask = os.umask(FILE_UMASK)
        with open(file_path, 'wb') as f:
            f.write(response.content)
        os.umask(previous_umask)

        return saved_filename



import logging

from djpsa.halo import models
from djpsa.halo.records import api
from djpsa.halo import sync


logger = logging.getLogger(__name__)


class BudgetDataSynchronizer(sync.HaloSynchronizer):
    model_class = models.BudgetDataTracker
    client_class = api.BudgetDataAPI

    related_meta = {
        'ticket_id': (models.Ticket, 'ticket'),
        'budgettype_id': (models.BudgetType, 'budget_type'),
    }

    def fetch_records(self, results, params=None):
        batch = 1
        for ticket in models.Ticket.objects.all().values_list(
                'id', flat=True):
            logger.info(
                'Fetching {} records, batch {}'.format(
                    self.get_model_name(), batch)
            )
            response = self.client.get(ticket)
            records = self._unpack_records(response)
            self.persist_page(records, results)
            batch += 1

        return results

    def _unpack_records(self, response):
        return response.get('budgets', list())

    @property
    def object_ids(self):
        object_ids = models.Ticket.objects.all() \
            .values_list('id', flat=True)

        return object_ids

    def _assign_field_data(self, instance, json_data):
        instance.id = json_data.get('id')
        instance.hours = json_data.get('hours')
        instance.rate = json_data.get('rate')
        instance.money = json_data.get('money')
        instance.estimated_hours = json_data.get('estimated_hours')
        instance.estimated_money = json_data.get('estimated_money')
        instance.actual_hours = json_data.get('actual_hours')
        instance.actual_money = json_data.get('actual_money')
        instance.scheduled_hours = json_data.get('scheduled_hours')
        instance.scheduled_value = json_data.get('scheduled_value')
        instance.toschedule_hours = json_data.get('toschedule_hours')
        instance.toschedule_value = json_data.get('toschedule_value')
        instance.remaining_hours = json_data.get('remaining_hours')
        instance.remaining_value = json_data.get('remaining_value')

        self.set_relations(instance, json_data)



from collections import OrderedDict

from djpsa.halo.records import sync
from djpsa.sync.management.commands.base_sync import BaseSyncCommand


class Command(BaseSyncCommand):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.synchronizer_map = OrderedDict(sync.sync_command_list)



# Generated by Django 4.2.11 on 2024-12-13 15:33

from django.db import migrations, models
import django.db.models.deletion


def nullify_team_field(apps, schema_editor):
    Ticket = apps.get_model('halo', 'Ticket')
    Ticket.objects.update(team=None)


class Migration(migrations.Migration):

    dependencies = [
        ('halo', '0018_agent_teams'),
    ]

    operations = [
        migrations.RunPython(nullify_team_field, reverse_code=nullify_team_field),
        migrations.AlterField(
            model_name='ticket',
            name='team',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='halo.team'),
        ),
    ]


from unittest import TestCase
from django.utils import timezone
from dateutil.parser import parse
from djpsa.halo.sync import empty_date_parser, ResponseKeyMixin


class TestEmptyDateParser(TestCase):

    def test_valid_date(self):
        date_str = "2023-10-10T10:00:00"
        expected_date = timezone.make_aware(parse(date_str), timezone.utc)
        self.assertEqual(empty_date_parser(date_str), expected_date)

    def test_empty_date(self):
        date_str = "1900-01-01T00:00:00"
        self.assertIsNone(empty_date_parser(date_str))

    def test_none_date(self):
        self.assertIsNone(empty_date_parser(None))

    def test_edge_case_date(self):
        date_str = "1980-01-01T00:00:00"
        self.assertIsNone(empty_date_parser(date_str))

    def test_above_edge_case_date(self):
        date_str = "1981-01-01T00:00:00"
        expected_date = timezone.make_aware(parse(date_str), timezone.utc)
        self.assertEqual(empty_date_parser(date_str), expected_date)


class TestResponseKeyMixin(TestCase):

    def setUp(self):
        class TestClass(ResponseKeyMixin):
            response_key = 'data'

        self.test_instance = TestClass()

    def test_unpack_records(self):
        response = {'data': [{'id': 1, 'name': 'Test'}]}
        expected_records = [{'id': 1, 'name': 'Test'}]
        self.assertEqual(
            self.test_instance._unpack_records(response), expected_records)



import logging
import requests
import hashlib
from django.conf import settings
from django.core.cache import cache

from djpsa.api.client import APIClient
from djpsa.api.exceptions import APIError, APIClientError
from djpsa.utils import LockNotAcquiredError, redis_lock

logger = logging.getLogger(__name__)


HALO_TOKEN_CACHE_NAME = 'halo_token:{}:{}'
CACHE_EXPIRE_TIME = 3540  # 1 hour less 1 minute so the token expires
# locally before expiring remotely, so we avoid access denied errors.
TOKEN_REQUEST_TIMEOUT = 30
TOKEN_LOCK_LIFETIME = TOKEN_REQUEST_TIMEOUT + 1
TOKEN_LOCK_ACQUIRE_TIMEOUT = 60


class HaloAPICredentials:
    def __init__(self, authorisation_server, client_id, client_secret):
        self.authorisation_server = authorisation_server
        if not self.authorisation_server.endswith('/'):
            self.authorisation_server += '/'
        self.client_id = client_id
        self.client_secret = client_secret


class HaloAPIClient(APIClient):
    endpoint = None

    def __init__(self,
                 conditions=None,
                 credentials=None,
                 resource_server=None,
                 token_locking=True):
        super().__init__(conditions)
        if not credentials:
            credentials = HaloAPICredentials(
                settings.HALO_AUTHORISATION_SERVER,
                settings.HALO_CLIENT_ID,
                settings.HALO_CLIENT_SECRET,
            )

        self.resource_server = resource_server or settings.HALO_RESOURCE_SERVER
        if not self.resource_server.endswith('/'):
            self.resource_server += '/'

        self.token_fetcher = HaloAPITokenFetcher(credentials, token_locking)

    def check_auth(self):
        return bool(self.token_fetcher.get_token(use_cache=False))

    def get_page(self, page=None, batch_size=None, params=None):
        params = params or {}
        if page:
            params['page_no'] = page
        if batch_size:
            params['page_size'] = batch_size
        return self.fetch_resource(params=params)

    def get(self, record_id):
        return self.request('GET', params={'search_id': record_id})

    def create(self, data):
        # Halo API expects a list of records, even if we're only creating one
        if not isinstance(data, list):
            data = [data]

        return self.request('POST', body=data)

    def update(self, record_id, data):
        data.update({'id': record_id})
        if not isinstance(data, list):
            data = [data]
        return self.request('POST', body=data)

    def delete(self, record_id):
        return self.request(
            'DELETE',
            endpoint_url=self._format_endpoint(record_id)
        )

    def remove_condition(self, condition):
        # self.conditions is a list of dictionaries, so we need to remove the
        # dictionary that contains the key we want to remove, value doesn't
        # matter
        self.conditions = \
            [c for c in self.conditions if c.get(condition) is None]

    def _format_endpoint(self, record_id=None):

        if record_id:
            return '{}{}/{}'.format(
                self.resource_server, self.endpoint, record_id)

        return '{}{}'.format(self.resource_server, self.endpoint)

    def _format_params(self, params=None):
        params = params or {}
        request_params = {}

        for condition in self.conditions:
            request_params.update(condition)

        request_params.update(params)

        if 'page_no' in request_params:
            request_params['page_size'] = \
                params.get('page_size', self.request_settings['batch_size'])
            request_params['pageinate'] = True

        return request_params

    def _request(
            self, method, endpoint_url, headers=None, params=None, **kwargs):
        token = self.token_fetcher.get_token()
        token_header = {'Authorization': f'Bearer {token}'}

        # If kwargs contains headers, update it with the token, if not,
        # create it
        if headers:
            headers.update(token_header)
        else:
            headers = token_header

        logger.debug(
            'Making %s request to %s: params %s, kwargs %s',
            method, endpoint_url, params, kwargs
        )
        # Make the actual request
        response = requests.request(
            method,
            endpoint_url,
            headers=headers,
            params=params,
            timeout=self.request_settings['timeout'],
            **kwargs
        )

        # If the token is invalid, refresh it and retry
        if response.status_code == 401:
            token = self.token_fetcher.get_token()
            headers['Authorization'] = f'Bearer {token}'
            response = requests.request(
                method,
                endpoint_url,
                headers=headers,
                params=params,
                **kwargs
            )

        return response

    def _prepare_error_response(self, response):
        result = response.json()
        error = ""

        if isinstance(result, str):
            # Error responses can be a string.
            return result

        try:
            error = result['error']

            if isinstance(error, str):
                # Error in some cases can just be a string.
                error_msg = f"Error: {error}"
                logger.error(error_msg)
                return error_msg

            error_desc = error['error_description']

        except KeyError:
            try:
                error = result['ClassName']
                error_desc = result['Message']
            except KeyError:
                try:
                    if len(result) == 1:
                        # This is about the best we can do. The Halo API
                        # doesn't provide a standard error format. There's
                        # no telling how deep the rabbit hole goes. We will
                        # just have to handle new error types as they come.
                        error_desc = (list(result.values())[0]
                                      .replace(
                                          "\r", ""
                                      ).replace(
                                          "\n", ""
                                      ).replace(
                                          "\'", "")
                                      )
                    else:
                        logger.error(f"Unknown error format: {result}")
                        error_desc = "An unknown error has occurred."
                except Exception as e:
                    logger.error(
                        f"Failed to process error from response: "
                        f"{result}, {e}"
                    )
                    raise e
            except TypeError:
                logger.error(f"Unknown error format: {result}")
                error_desc = "An unknown error has occurred."

        return f"{error}: {error_desc}" if error else error_desc


class WebhookAPIClient(HaloAPIClient):
    endpoint = 'webhook'

    def get(self, **kwargs):
        return self.request('GET')


class NotificationAPIClient(HaloAPIClient):
    # API for webhook event creation, tickets, clients, etc
    endpoint = 'notification'


class HaloAPITokenFetcher:
    def __init__(self, credentials, token_locking=True):
        self.credentials = credentials
        self.token_locking = token_locking

    def get_token(self, use_cache=True):
        if use_cache:
            token = self._get_saved_token()
            if token:
                return token

        if self.token_locking:
            try:
                with redis_lock(
                        'halo_token_lock',
                        TOKEN_LOCK_LIFETIME,
                        TOKEN_LOCK_ACQUIRE_TIMEOUT):
                    # Check if a valid token is already available. May
                    # have been fetched by another worker thread while we
                    # were waiting.
                    if use_cache:
                        token = self._get_saved_token()
                        if token:
                            return token

                    # No valid token- get a new one.
                    # We'll save it even if we didn't use the cache, because
                    # it's possible getting a new token invalidates prior
                    # tokens.
                    return self._get_new_token_and_save()
            except LockNotAcquiredError as e:
                logger.error(f"Could not acquire lock: {e}")
        else:
            # Get a new token without locking
            return self._get_new_token_and_save()

    def _get_cache_name(self):
        """Return the cache name for the token."""
        # One-way hash of secret so it's not stored in plain text
        secret_hash = hashlib.sha256(
            self.credentials.client_secret.encode('utf-8')
        ).hexdigest()
        return HALO_TOKEN_CACHE_NAME.format(
            self.credentials.client_id, secret_hash)

    def _get_saved_token(self):
        return cache.get(self._get_cache_name())

    def _get_new_token_and_save(self):
        logger.debug('Getting new access token')
        token_url = '{}token'.format(self.credentials.authorisation_server)
        try:
            response = requests.post(
                token_url,
                data={
                    'grant_type': 'client_credentials',
                    'client_id': self.credentials.client_id,
                    'client_secret': self.credentials.client_secret,
                    'scope': 'all',
                },
                timeout=TOKEN_REQUEST_TIMEOUT,
            )
            response.raise_for_status()
            token = response.json()['access_token']
        except requests.RequestException as e:
            logger.error(f"Failed to get new token: {e}")
            raise_cls = APIError
            if e.response is not None and 400 <= e.response.status_code < 500:
                raise_cls = APIClientError
            raise raise_cls('{}'.format(e))

        cache.set(self._get_cache_name(), token, CACHE_EXPIRE_TIME)
        return token



# Generated by Django 4.2.11 on 2024-09-24 09:59

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('halo', '0011_tickettype_remove_project_agent_and_more'),
    ]

    operations = [
        migrations.AlterField(
            model_name='client',
            name='main_site',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='client_primary', to='halo.site'),
        ),
    ]



from typing import Any, List

from djpsa.halo import models
from djpsa.halo.records import api
from djpsa.halo import sync


class ClientSynchronizer(sync.ResponseKeyMixin, sync.HaloSynchronizer):
    response_key = 'clients'
    model_class = models.ClientTracker
    client_class = api.ClientAPI

    related_meta = {
        'main_site_id': (models.Site, 'site'),
    }

    def __init__(self,
                 full: bool = False,
                 conditions: List = None,
                 *args: Any,
                 **kwargs: Any):
        super().__init__(full, conditions, *args, **kwargs)

        self.client.add_condition({
            'includeactive': True,
        })

    def _assign_field_data(self, instance, json_data):

        instance.id = json_data.get('id')
        instance.name = json_data.get('name')
        instance.inactive = json_data.get('inactive')
        instance.phone_number = json_data.get('main_phonenumber')

        self.set_relations(instance, json_data)



from django.contrib import admin

from djpsa.halo import models


@admin.register(models.Ticket)
class TicketAdmin(admin.ModelAdmin):
    list_display = (
        'id',
        'summary',
        'client',
        'status',
        'priority',
        'agent',
        'type'
    )
    search_fields = [
        'id',
        'summary',
        'client__name',
        'status__name',
        'priority__name',
        'agent__name'
    ]
    list_filter = ('status', 'priority', 'agent', 'type')


@admin.register(models.Client)
class ClientAdmin(admin.ModelAdmin):
    list_display = ('id', 'name', 'inactive')
    search_fields = ['id', 'name']


@admin.register(models.Agent)
class AgentAdmin(admin.ModelAdmin):
    list_display = ('id', 'name', 'email', 'is_disabled')
    search_fields = ['id', 'name']


@admin.register(models.Status)
class StatusAdmin(admin.ModelAdmin):
    list_display = ('id', 'name')
    search_fields = ['id', 'name']


@admin.register(models.Priority)
class PriorityAdmin(admin.ModelAdmin):
    list_display = ('id', 'name', 'colour')
    search_fields = ['id', 'name']
    list_filter = ('is_hidden',)


@admin.register(models.TicketType)
class TicketTypeAdmin(admin.ModelAdmin):
    list_display = ('id', 'name')
    search_fields = ['id', 'name']


@admin.register(models.HaloUser)
class HaloUserAdmin(admin.ModelAdmin):
    list_display = ('id', 'name', 'email', 'active', 'agent', 'client', 'site')
    search_fields = ['id', 'name', 'email']


@admin.register(models.Appointment)
class AppointmentAdmin(admin.ModelAdmin):
    list_display = ('id', 'subject', 'start_date', 'end_date')
    search_fields = ['id', 'subject']


@admin.register(models.SLA)
class SLAAdmin(admin.ModelAdmin):
    list_display = ('id', 'name')
    search_fields = ['id', 'name']


@admin.register(models.Site)
class SiteAdmin(admin.ModelAdmin):
    list_display = ('id', 'name')
    search_fields = ['id', 'name']


@admin.register(models.Action)
class ActionAdmin(admin.ModelAdmin):
    list_display = (
        'id',
        'outcome',
        'action_arrival_date',
        'action_completion_date',
        'time_taken',
        'note')
    search_fields = \
        ['id', 'ticket__id', 'time_taken', 'project__id', 'agent__id']


@admin.register(models.Outcome)
class OutcomeAdmin(admin.ModelAdmin):
    list_display = ('id', 'outcome')
    search_fields = ['id', 'name']


@admin.register(models.Team)
class TeamAdmin(admin.ModelAdmin):
    list_display = ('id', 'name', 'ticket_count')
    search_fields = ['id', 'name']


@admin.register(models.ChargeRate)
class ChargeRateAdmin(admin.ModelAdmin):
    list_display = ('id', 'name')
    search_fields = ['id', 'name']


@admin.register(models.BudgetType)
class BudgetTypeAdmin(admin.ModelAdmin):
    list_display = ('id', 'name')
    search_fields = ['id', 'name']


@admin.register(models.BudgetData)
class BudgetDataAdmin(admin.ModelAdmin):
    list_display = ('id', 'budget_type', 'ticket')
    search_fields = ['id', 'budget_type__name', 'ticket__id']



# Generated by Django 4.2.11 on 2024-10-21 12:39

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('halo', '0015_alter_ticket_details_alter_ticket_email_to_list_and_more'),
    ]

    operations = [
        migrations.AlterField(
            model_name='appointment',
            name='site',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='halo.site'),
        ),
        migrations.AlterField(
            model_name='client',
            name='main_site',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='client_primary', to='halo.site'),
        ),
        migrations.AlterField(
            model_name='halouser',
            name='site',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='halo.site'),
        ),
        migrations.AlterField(
            model_name='site',
            name='sla',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='halo.sla'),
        ),
    ]



# Generated by Django 4.2.11 on 2025-02-10 14:49

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('halo', '0022_alter_budgetdata_ticket'),
    ]

    operations = [
        migrations.AddField(
            model_name='action',
            name='non_billable_time',
            field=models.FloatField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='action',
            name='outcome',
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
        migrations.AddField(
            model_name='action',
            name='travel_time',
            field=models.FloatField(blank=True, null=True),
        ),
    ]



from unittest import TestCase
from unittest.mock import MagicMock, patch
from django.conf import settings
from django.utils import timezone
from django.db import IntegrityError

from djpsa.sync.sync import Synchronizer, SyncResults, CREATED, UPDATED, \
    SKIPPED


class TestSynchronizer(TestCase):

    def setUp(self):
        # Mock settings
        settings.PROVIDER = MagicMock()
        settings.PROVIDER.get_request_settings.return_value = \
            {'batch_size': 100}

        # Mock model class
        self.model_class_mock = MagicMock()
        self.model_class_mock.objects.all.return_value = [1, 2, 3]

        # Mock client class
        self.client_class_mock = MagicMock()
        self.client_class_mock.return_value.get_page.\
            return_value = {'data': [{'id': 1}, {'id': 2}]}

        class MockSynchronizer(Synchronizer):
            model_class = self.model_class_mock
            client_class = self.client_class_mock
            lookup_key = 'id'

            def get_model_name(self):
                return 'MockModel'

        # Create an instance of Synchronizer
        self.synchronizer = MockSynchronizer()

    @patch('djpsa.sync.sync.SyncJob')
    def test_get_last_sync_job_time(self, _):

        with patch.object(
                self.synchronizer,
                '_format_job_condition') as format_job_condition_mock:
            # Mock SyncJob queryset
            sync_job_qset_mock = MagicMock()
            sync_job_qset_mock.count.return_value = 2
            last_sync_job = MagicMock()
            last_sync_job.start_time = timezone.now()
            sync_job_qset_mock.last.return_value = last_sync_job
            self.synchronizer.last_updated_field = 'updated_at'
            self.synchronizer.full = False
            self.synchronizer.partial_sync_support = True

            # Call the method
            self.synchronizer._get_last_sync_job_time(sync_job_qset_mock)

        # Assertions
        self.assertEqual(format_job_condition_mock.call_count, 1)

    @patch('djpsa.sync.sync.SyncJob')
    def test_get_last_sync_job_time_no_last_updated_field(self, _):
        # Mock SyncJob queryset
        sync_job_qset_mock = MagicMock()
        sync_job_qset_mock.count.return_value = 2
        self.synchronizer.last_updated_field = None
        self.synchronizer.full = False
        self.synchronizer.partial_sync_support = True

        # Call the method
        last_sync_time = \
            self.synchronizer._get_last_sync_job_time(sync_job_qset_mock)

        # Assertions
        self.assertIsNone(last_sync_time)

    @patch('djpsa.sync.sync.SyncJob')
    def test_get_last_sync_job_time_full_sync(self, _):
        # Mock SyncJob queryset
        sync_job_qset_mock = MagicMock()
        sync_job_qset_mock.count.return_value = 2
        self.synchronizer.last_updated_field = 'updated_at'
        self.synchronizer.full = True
        self.synchronizer.partial_sync_support = True

        # Call the method
        last_sync_time = \
            self.synchronizer._get_last_sync_job_time(sync_job_qset_mock)

        # Assertions
        self.assertIsNone(last_sync_time)

    @patch('djpsa.sync.sync.SyncJob')
    def test_get_last_sync_job_time_partial_sync_not_supported(self, _):
        # Mock SyncJob queryset
        sync_job_qset_mock = MagicMock()
        sync_job_qset_mock.count.return_value = 2
        self.synchronizer.last_updated_field = 'updated_at'
        self.synchronizer.full = False
        self.synchronizer.partial_sync_support = False

        # Call the method
        last_sync_time = \
            self.synchronizer._get_last_sync_job_time(sync_job_qset_mock)

        # Assertions
        self.assertIsNone(last_sync_time)

    @patch.object(Synchronizer, 'persist_page')
    @patch.object(Synchronizer, '_unpack_records')
    def test_fetch_records(self, mock_unpack_records, mock_persist_page):
        with patch.object(
                self.synchronizer.client_class, 'get_page') as mock_get_page:
            results = SyncResults()

            # First page response
            # Second page response
            # Empty response to stop the loop
            mock_get_page.side_effect = [
                {'data': 'page1'},
                {'data': 'page2'},
                {}
            ]

            # Unpacked records from first page
            # Unpacked records from second page
            # No records in the last page
            mock_unpack_records.side_effect = [
                [{'id': 1, 'summary': 'record1'},
                 {'id': 2, 'summary': 'record2'}],
                [{'id': 3, 'summary': 'record3'}, ],
                []
            ]

            self.synchronizer.batch_size = 2

            # Execute
            results = self.synchronizer.fetch_records(results)

            self.synchronizer.batch_size = 100

            # Verify
            self.assertEqual(mock_unpack_records.call_count, 2)
            self.assertEqual(mock_persist_page.call_count, 2)
            self.assertEqual(results.created_count, 0)
            self.assertEqual(results.updated_count, 0)
            self.assertEqual(results.skipped_count, 0)
            self.assertEqual(results.deleted_count, 0)

    def test_persist_page_created(self):
        records = [{'id': 1}, {'id': 2}]
        results = SyncResults()
        self.synchronizer.update_or_create_instance = MagicMock(side_effect=[
            (MagicMock(), CREATED),
            (MagicMock(), CREATED)
        ])

        self.synchronizer.persist_page(records, results)

        self.assertEqual(results.created_count, 2)
        self.assertEqual(results.updated_count, 0)
        self.assertEqual(results.skipped_count, 0)
        self.assertEqual(results.synced_ids, {1, 2})

    def test_persist_page_updated(self):
        records = [{'id': 1}, {'id': 2}]
        results = SyncResults()
        self.synchronizer.update_or_create_instance = MagicMock(side_effect=[
            (MagicMock(), UPDATED),
            (MagicMock(), UPDATED)
        ])

        self.synchronizer.persist_page(records, results)

        self.assertEqual(results.created_count, 0)
        self.assertEqual(results.updated_count, 2)
        self.assertEqual(results.skipped_count, 0)
        self.assertEqual(results.synced_ids, {1, 2})

    def test_persist_page_skipped(self):
        records = [{'id': 1}, {'id': 2}]
        results = SyncResults()
        self.synchronizer.update_or_create_instance = MagicMock(side_effect=[
            (MagicMock(), SKIPPED),
            (MagicMock(), SKIPPED)
        ])

        self.synchronizer.persist_page(records, results)

        self.assertEqual(results.created_count, 0)
        self.assertEqual(results.updated_count, 0)
        self.assertEqual(results.skipped_count, 2)
        self.assertEqual(results.synced_ids, {1, 2})

    @patch('djpsa.sync.sync.logger')
    def test_persist_page_exception(self, mock_logger):
        records = [{'id': 1}, {'id': 2}]
        results = SyncResults()
        self.synchronizer.update_or_create_instance = \
            MagicMock(side_effect=IntegrityError('Test error'))

        self.synchronizer.persist_page(records, results)

        self.assertEqual(results.created_count, 0)
        self.assertEqual(results.updated_count, 0)
        self.assertEqual(results.skipped_count, 0)
        self.assertEqual(results.synced_ids, {1, 2})
        self.assertTrue(mock_logger.warning.called)

    def test_prune_stale_records(self):
        mock_get_delete_qset = self.synchronizer.get_delete_qset = MagicMock()
        initial_ids = {1, 2, 3, 4}
        synced_ids = {2, 3}

        mock_delete_qset = MagicMock()
        mock_delete_qset.count.return_value = 2
        mock_get_delete_qset.return_value = mock_delete_qset

        deleted_count = \
            self.synchronizer.prune_stale_records(initial_ids, synced_ids)

        self.assertEqual(deleted_count, 2)
        mock_get_delete_qset.assert_called_once_with({1, 4})
        mock_delete_qset.delete.assert_called_once()



# Generated by Django 4.2.16 on 2026-01-30 14:26

from django.db import migrations
import django.utils.timezone
import django_extensions.db.fields


class Migration(migrations.Migration):

    dependencies = [
        ('halo', '0029_alter_ticket_appointment_type'),
    ]

    operations = [
        migrations.AddField(
            model_name='ticket',
            name='created',
            field=django_extensions.db.fields.CreationDateTimeField(auto_now_add=True, default=django.utils.timezone.now, verbose_name='created'),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='ticket',
            name='modified',
            field=django_extensions.db.fields.ModificationDateTimeField(auto_now=True, verbose_name='modified'),
        ),
    ]



from django.db import models
from model_utils import FieldTracker


class Site(models.Model):

    name = models.CharField(max_length=255)
    client = models.ForeignKey(
        'Client', blank=True, null=True, on_delete=models.CASCADE)
    colour = models.CharField(max_length=50, null=True, blank=True)
    active = models.BooleanField(default=True)
    phone_number = models.CharField(blank=True, null=True, max_length=250)
    sla = models.ForeignKey(
        'SLA', blank=True, null=True, on_delete=models.SET_NULL)
    use = models.CharField(max_length=255, blank=True, null=True)
    delivery_address = models.CharField(max_length=1000, blank=True, null=True)

    class Meta:
        verbose_name_plural = "Sites"

    def __str__(self):
        return str(self.name)


class SiteTracker(Site):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'halo_site'



from djpsa.halo import models
from djpsa.halo.records import api

from djpsa.halo import sync


class TeamSynchronizer(sync.HaloSynchronizer):
    model_class = models.TeamTracker
    client_class = api.TeamAPI

    def _assign_field_data(self, instance, json_data):

        instance.id = json_data.get('id')
        instance.name = json_data.get('name')
        instance.ticket_count = json_data.get('ticket_count')



from djpsa.halo import models
from djpsa.halo.records import api
from djpsa.halo import sync


class OutcomeSynchronizer(sync.HaloSynchronizer):
    model_class = models.OutcomeTracker
    client_class = api.OutcomeAPI

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.client.add_condition({
            'showhidden': True,
        })

    def _assign_field_data(self, instance, json_data):
        instance.id = json_data.get('id')
        instance.outcome = json_data.get('outcome')
        instance.button_name = json_data.get('buttonname')
        instance.label_long = json_data.get('labellong')
        instance.sequence = json_data.get('sequence')
        instance.hidden = json_data.get('hidden', False)
        instance.icon = json_data.get('icon')



from django.db import models
from model_utils import FieldTracker


class Outcome(models.Model):
    """
    Outcomes are essentially "Action Types"
    """
    # outcome is the name field of Outcome
    outcome = models.CharField(max_length=255)
    button_name = models.CharField(max_length=255)
    label_long = models.CharField(max_length=255)
    sequence = models.IntegerField()
    hidden = models.BooleanField(default=False)
    icon = models.CharField(max_length=255)

    def __str__(self):
        return str(self.outcome)


class OutcomeTracker(Outcome):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'halo_outcome'



from django.db import models
from model_utils import FieldTracker

# This is not a real record in Halo, but created for the purpose
# of easily storing and processing budget data. Uses the Ticket
# API to get a tickets budget data instead of having its own.


class BudgetData(models.Model):
    ticket = models.ForeignKey(
        'Ticket', on_delete=models.CASCADE)
    budget_type = models.ForeignKey(
        'BudgetType', on_delete=models.CASCADE, blank=True, null=True)
    hours = models.DecimalField(
        max_digits=10, decimal_places=2, blank=True, null=True)
    rate = models.DecimalField(
        max_digits=10, decimal_places=2, blank=True, null=True)
    money = models.DecimalField(
        max_digits=10, decimal_places=2, blank=True, null=True)
    estimated_hours = models.DecimalField(
        max_digits=10, decimal_places=2, blank=True, null=True)
    estimated_money = models.DecimalField(
        max_digits=10, decimal_places=2, blank=True, null=True)
    actual_hours = models.DecimalField(
        max_digits=10, decimal_places=2, blank=True, null=True)
    actual_money = models.DecimalField(
        max_digits=10, decimal_places=2, blank=True, null=True)
    scheduled_hours = models.DecimalField(
        max_digits=10, decimal_places=2, blank=True, null=True)
    scheduled_value = models.DecimalField(
        max_digits=10, decimal_places=2, blank=True, null=True)
    toschedule_hours = models.DecimalField(
        max_digits=10, decimal_places=2, blank=True, null=True)
    toschedule_value = models.DecimalField(
        max_digits=10, decimal_places=2, blank=True, null=True)
    remaining_hours = models.DecimalField(
        max_digits=10, decimal_places=2, blank=True, null=True)
    remaining_value = models.DecimalField(
        max_digits=10, decimal_places=2, blank=True, null=True)

    class Meta:
        verbose_name_plural = "Budget Data"


class BudgetDataTracker(BudgetData):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'halo_budgetdata'



from djpsa.halo import models
from djpsa.halo.records import api
from djpsa.halo import sync
from djpsa.halo.records.halouser.model import HaloUser


class ActionSynchronizer(sync.ResponseKeyMixin,
                         sync.HaloChildFetchRecordsMixin,
                         sync.CreateMixin,
                         sync.UpdateMixin,
                         sync.HaloSynchronizer,
                         ):
    lookup_key = 'ticket_action_id'
    model_class = models.ActionTracker
    client_class = api.ActionAPI
    response_key = 'actions'
    parent_field = 'ticket_id'
    parent_model_class = models.Ticket

    related_meta = {
        'ticket_id': (models.TicketTracker, 'ticket'),
        'project_id': (models.TicketTracker, 'project'),
        'who_agentid': (models.Agent, 'agent'),
        'outcome_id': (models.Outcome, 'outcome'),
        'charge_rate_id': (models.ChargeRate, 'chargerate'),
    }

    def update_or_create_instance(self, json_data):
        # So far, action is the only model that has to be treated in this way,
        #  so rather than updating the lookup_key pattern to be more generic,
        #  we'll just add this special case here. As I think it would just add
        #  confusion if there is such a generic pattern for one model, and
        #  EVERYTHING else.

        # Action IDs are only unique per ticket, so concatenate the ticket ID
        # with the action ID
        json_data[self.lookup_key] = \
            int(f"{json_data.get('ticket_id')}{json_data.get('id')}")

        return super().update_or_create_instance(json_data)

    def _assign_field_data(self, instance, json_data):
        instance.id = json_data.get(self.lookup_key)
        instance.action_arrival_date = \
            sync.empty_date_parser(json_data.get('actionarrivaldate'))
        instance.action_completion_date = \
            sync.empty_date_parser(json_data.get('actioncompletiondate'))
        instance.action_date_created = \
            sync.empty_date_parser(json_data.get('actiondatecreated'))
        instance.time_taken = json_data.get('timetaken')
        instance.time_taken_adjusted = json_data.get('timetakenadjusted')
        instance.time_taken_days = json_data.get('timetakendays')
        instance.non_billable_time = json_data.get('nonbilltime')
        instance.travel_time = json_data.get('traveltime')
        instance.note = json_data.get('note')
        instance.action_charge_amount = json_data.get('actionchargeamount')
        instance.action_charge_hours = json_data.get('actionchargehours')
        instance.action_non_charge_amount = \
            json_data.get('actionnonchargeamount')
        instance.action_non_charge_hours = \
            json_data.get('actionnonchargehours')
        instance.attachment_count = json_data.get('attachment_count')
        instance.act_is_billable = json_data.get('actisbillable', False)
        instance.hidden_from_user = json_data.get('hiddenfromuser', False)
        instance.important = json_data.get('important', False)

        self.set_relations(instance, json_data)

    def create(self, data, *args, **kwargs):
        # Halo impersonation is a bit weird, we need to get the agent's
        # username from the user record related to the agent. Also, the
        # username field is called "name".
        data['agent'] = HaloUser.objects.get(agent=data['agent']).name

        return super().create(data, *args, **kwargs)




from django.urls import re_path

from djpsa.halo import views

app_name = 'djpsa.halo'


urlpatterns = [
    re_path(
        r'^ticket/$',
        view=views.TicketCallBackView.as_view(),
        name='ticket_callback'
    ),
]



# Generated by Django 4.2.11 on 2024-10-15 13:17

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('halo', '0012_alter_client_main_site'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='ticket',
            name='start_time',
        ),
        migrations.RemoveField(
            model_name='ticket',
            name='target_time',
        ),
        migrations.CreateModel(
            name='Action',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('action_arrival_date', models.DateTimeField(blank=True, null=True)),
                ('action_completion_date', models.DateTimeField(blank=True, null=True)),
                ('action_date_created', models.DateTimeField(blank=True, null=True)),
                ('time_taken', models.FloatField(blank=True, null=True)),
                ('time_taken_adjusted', models.FloatField(blank=True, null=True)),
                ('time_taken_days', models.FloatField(blank=True, null=True)),
                ('note', models.TextField(blank=True, null=True)),
                ('action_charge_amount', models.FloatField(blank=True, null=True)),
                ('action_charge_hours', models.FloatField(blank=True, null=True)),
                ('action_non_charge_amount', models.FloatField(blank=True, null=True)),
                ('action_non_charge_hours', models.FloatField(blank=True, null=True)),
                ('act_is_billable', models.BooleanField(default=False)),
                ('attachment_count', models.IntegerField(blank=True, null=True)),
                ('charge_rate', models.IntegerField(blank=True, null=True)),
                ('hidden_from_user', models.BooleanField(default=False)),
                ('important', models.BooleanField(default=False)),
                ('agent', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='halo.agent')),
                ('project', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='halo.ticket')),
                ('ticket', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='halo.ticket')),
            ],
            options={
                'verbose_name_plural': 'Actions',
            },
        ),
        migrations.CreateModel(
            name='ActionTracker',
            fields=[
            ],
            options={
                'db_table': 'halo_action',
                'proxy': True,
                'indexes': [],
                'constraints': [],
            },
            bases=('halo.action',),
        ),
    ]




class APIError(Exception):
    """Raise this, not request exceptions."""
    pass


class APIClientError(APIError):
    """
    Raise this to indicate any http error that falls within the
    4xx class of http status codes.
    """
    pass


class APIServerError(APIError):
    """
    Raise this to indicate a Server Error
    """
    pass


class RecordNotFoundError(APIClientError):
    """The record was not found."""
    pass


class SecurityPermissionsException(APIClientError):
    """The API credentials have insufficient security permissions."""
    pass



import logging

from typing import List, Any
from django.utils import timezone
from django.db import transaction, IntegrityError
from django.conf import settings

from djpsa.sync.models import SyncJob
from djpsa.utils import get_djpsa_settings

logger = logging.getLogger(__name__)

CREATED = 1
UPDATED = 2
SKIPPED = 3


def log_sync_job(f):
    def wrapper(*args, **kwargs):
        sync_instance = args[0]
        created_count = updated_count = deleted_count = skipped_count = 0
        sync_job = SyncJob()
        sync_job.start_time = timezone.now()
        sync_job.entity_name = sync_instance.get_model_name()
        sync_job.synchronizer_class = \
            sync_instance.__class__.__name__

        if sync_instance.full:
            sync_job.sync_type = 'full'
        else:
            sync_job.sync_type = 'partial'

        sync_job.save()

        try:
            created_count, updated_count, skipped_count, deleted_count = \
                f(*args, **kwargs)
            sync_job.success = True
        except Exception as e:
            sync_job.message = str(e.args[0])
            sync_job.success = False
            raise
        finally:
            sync_job.end_time = timezone.now()
            sync_job.added = created_count
            sync_job.updated = updated_count
            sync_job.skipped = skipped_count
            sync_job.deleted = deleted_count
            sync_job.save()

        return created_count, updated_count, skipped_count, deleted_count

    return wrapper


class InvalidObjectException(Exception):
    """
    If for any reason an object can't be created (for example, it references
    an unknown foreign object, or is missing a required field), raise this
    so that the synchronizer can catch it and continue with other records.
    """
    pass


class SyncResults:
    """Track results of a sync job."""

    def __init__(self):
        self.created_count = 0
        self.updated_count = 0
        self.skipped_count = 0
        self.deleted_count = 0
        self.synced_ids = set()


class Synchronizer:
    lookup_key = 'id'
    model_class = None
    client_class = None
    last_updated_field = None
    bulk_prune = True

    def __init__(self,
                 full: bool = False,
                 conditions: List = None,
                 *args: Any,
                 **kwargs: Any):

        self.sync_settings = get_djpsa_settings()
        if hasattr(settings, 'DJPSA_CONF_CALLABLE'):
            self.sync_settings.update(
                settings.DJPSA_CONF_CALLABLE().get('sync', {}))

        conditions = conditions or []
        self.client = self.client_class(conditions)
        self.partial_sync_support = True
        self.batch_size = self.sync_settings['batch_size']
        self.full = full

    def get_sync_job_qset(self):
        return SyncJob.objects.filter(
            entity_name=self.get_model_name()
        )

    def get_model_name(self):
        return self.model_class.__bases__[0].__name__

    def _get_last_sync_job_time(self, sync_job_qset):
        if sync_job_qset.count() > 1 and self.last_updated_field \
                and not self.full and self.partial_sync_support:
            return self._format_job_condition(
                sync_job_qset.last().start_time.strftime(
                    '%Y-%m-%dT%H:%M:%S.%fZ'))
        return None

    @log_sync_job
    def sync(self):
        if not self.full:
            sync_job_qset = self.get_sync_job_qset().filter(success=True)

            last_sync_job_condition = \
                self._get_last_sync_job_time(sync_job_qset)

            if last_sync_job_condition:
                self.client.add_condition(last_sync_job_condition)

        results = SyncResults()

        # Set of IDs of all records prior to sync,
        # to find stale records for deletion.
        initial_ids = self.instance_ids() if self.full else []

        results = self.fetch_records(results)

        results = self._post_sync_operations(results)

        if self.full:
            results.deleted_count = self.prune_stale_records(
                initial_ids, results.synced_ids
            )

        return results.created_count, results.updated_count, \
            results.skipped_count, results.deleted_count

    def _post_sync_operations(self, results):
        return results

    def instance_ids(self, filter_params=None):
        ids = self.model_class.objects.all().order_by('id') \
            .values_list('id', flat=True)
        if filter_params:
            ids = ids.filter(**filter_params)
        return set(ids)

    def fetch_records(self, results, params=None):
        """
        For all pages of results, save each page of results to the DB.
        """
        page = 1
        last_recorded_id = None

        while True:
            logger.info(
                'Fetching {} records, batch {}'.format(
                    self.get_model_name(), page)
            )
            response = self.client.get_page(
                page=page, batch_size=self.batch_size, params=params)
            records = self._unpack_records(response)

            current_id = records[0]['id'] if records else None
            if last_recorded_id == current_id:
                # Halo has sent the same page of records again, so we've
                # reached the end of the records. (Other PSA APIs do
                # not have this issue.) Break now so we don't add extra
                # records to the skipped count.
                break

            self.persist_page(records, results)
            page += 1
            if len(records) < self.batch_size:
                # This page wasn't full, so there's no more records after
                # this page.
                break
            last_recorded_id = records[0]['id'] if records else None

        return results

    def persist_page(self, records, results):
        """Persist one page of records to DB."""
        for record in records:
            if not self._try_validate(record):
                # Skip this record, if it doesn't meet some criteria
                # defined in a child synchronizer. Do not count it
                # as a skipped record.
                continue

            try:
                with transaction.atomic():
                    instance, result = self.update_or_create_instance(record)
                if result == CREATED:
                    results.created_count += 1
                elif result == UPDATED:
                    results.updated_count += 1
                else:
                    results.skipped_count += 1
            except (IntegrityError, InvalidObjectException) as e:
                logger.warning('{}'.format(e))

            results.synced_ids.add(record[self.lookup_key])

        return results

    def update(self, record_id, data, *args, **kwargs):
        raise NotImplementedError(
            f"This synchronizer does not support updates: {self}")

    def create(self, data, *args, **kwargs):
        raise NotImplementedError(
            f"This synchronizer does not support creation: {self}")

    def delete(self, record_id, *args, **kwargs):
        raise NotImplementedError(
            f"This synchronizer does not support deletion: {self}")

    def set_relations(self, instance, json_data):
        """
        Set the foreign key relations on the instance as
        defined in the related_meta dictionary.
        """
        for json_field, value in self.related_meta.items():
            model_class, field_name = value
            self._assign_relation(
                instance,
                json_data,
                json_field,
                model_class,
                field_name
            )

    def _assign_field_data(self, instance, api_instance):
        """
        Assign the field data from the API instance to the model instance,
        overriding this method in the subclass to handle the specific fields.
        """
        raise NotImplementedError

    def _format_job_condition(self, last_sync_time):
        raise NotImplementedError

    def _unpack_records(self, response):
        """
        Unpack the records from the response and return them as a list.
        """
        # Override this method to unpack the records from the response, as
        # the structure of the response varies between different PSAs and
        # even between different endpoints in the same PSA.
        return response

    @staticmethod
    def _assign_null_relation(instance, model_field):
        """
        Set the FK to null, but handle issues like the FK being non-null.

        This can happen because PSAs can give us records that point to
        non-existent records- such as activities whose assignTo fields point
        to deleted members.
        """
        try:
            setattr(instance, model_field, None)
        except ValueError:
            # The model_field may have been non-null.
            raise InvalidObjectException(
                "Unable to set field {} on {} to null, as it's required.".
                format(model_field, instance)
            )

    def _assign_relation(self, instance, json_data,
                         json_field, model_class, model_field):
        """
        Look up the given foreign relation, and set it to the given
        field on the instance.
        """
        relation_id = json_data.get(json_field)
        if relation_id is None:
            self._assign_null_relation(instance, model_field)
            return

        uid = relation_id

        try:
            related_instance = model_class.objects.get(pk=uid)
            setattr(instance, model_field, related_instance)
        except model_class.DoesNotExist:
            logger.warning(
                'Failed to find {} {} for {} {}.'.format(
                    json_field,
                    uid,
                    type(instance),
                    instance.id
                )
            )
            self._assign_null_relation(instance, model_field)

    def _clean_data(self, data):
        """
        Clean the data from the API before using it to create/update
        the model instance.
        Override this method in child classes if needed.
        """

        # Remove any null bytes from string fields
        for key, value in data.items():
            if isinstance(value, str):
                data[key] = value.replace("\x00", "\uFFFD")

        return data

    def update_or_create_instance(self, api_instance):
        """
        Creates and returns an instance if it does not already exist.
        """
        result = None
        instance_pk = api_instance[self.lookup_key]

        try:
            instance = self.model_class.objects.get(pk=instance_pk)
        except self.model_class.DoesNotExist:
            instance = self.model_class()
            result = CREATED

        try:
            self._clean_data(api_instance)

            self._assign_field_data(instance, api_instance)

            if result == CREATED:
                try:
                    instance.save()
                except IntegrityError as e:
                    # Race condition: another process created this record
                    # between our get() and save(). Re-fetch and update.
                    try:
                        instance = self.model_class.objects.get(
                            pk=instance_pk)
                    except self.model_class.DoesNotExist:
                        # If bug described in #4412 still happens somehow,
                        # re-raise so we know.
                        raise e
                    self._assign_field_data(instance, api_instance)
                    if self._is_instance_changed(instance):
                        instance.save()
                        result = UPDATED
                    else:
                        result = SKIPPED
            elif self._is_instance_changed(instance):
                instance.save()
                result = UPDATED
            else:
                result = SKIPPED
        except AttributeError as e:
            msg = "AttributeError while attempting to sync object {}." \
                  " Error: {}".format(self.model_class, e)
            logger.error(msg)
            raise InvalidObjectException(msg)
        except IntegrityError as e:
            msg = "IntegrityError while attempting to create {}." \
                  " Error: {}".format(self.model_class, e)
            logger.error(msg)
            raise InvalidObjectException(msg)

        if result == CREATED:
            result_log = 'Created'
        elif result == UPDATED:
            result_log = 'Updated'
        else:
            result_log = 'Skipped'

        logger.info('{}: {} {}'.format(
            result_log,
            self.get_model_name(),
            instance
        ))

        return instance, result

    def prune_stale_records(self, initial_ids, synced_ids):
        """
        Delete records that existed when sync started but were
        not seen as we iterated through all records from REST API.

        If bulk_prune is set to False, delete records one by one to
        prevent errors.
        """
        stale_ids = initial_ids - synced_ids
        deleted_count = 0
        if stale_ids:
            delete_qset = self.get_delete_qset(stale_ids)
            deleted_count = delete_qset.count()

            logger.info(
                'Removing {} stale records for model: {}'.format(
                    len(stale_ids), self.get_model_name(),
                )
            )

            # If bulk_prune is set to True, delete records in bulk
            if self.bulk_prune:
                try:
                    delete_qset.delete()
                except IntegrityError as e:
                    logger.exception(
                        'IntegrityError while attempting to delete {} '
                        'records. Error: {}'.format(
                            self.get_model_name(), e)
                    )
            else:
                for instance in delete_qset:
                    try:
                        instance.delete()
                    except IntegrityError as e:
                        logger.exception(
                            'IntegrityError while attempting to delete {} '
                            'records. Error: {}'.format(
                                self.get_model_name(), e)
                        )

        return deleted_count

    def get_delete_qset(self, stale_ids):
        return self.model_class.objects.filter(pk__in=stale_ids)

    def _is_instance_changed(self, instance):
        return instance.tracker.changed()

    def _try_validate(self, record):
        """
        For some record types, filtering out records that don't meet
        certain criteria via the API may not be possible. In these cases,
        override this method to perform a check in child synchronizers to
        remove records that don't meet the child class' criteria.

        This method should return True if the record is valid, and False if it
        should be skipped. The reason we want to skip this record is that we
        don't want this record to be counted as a skipped record, or included
        in the results in any way. As if it was never requested in the first
        place.

        By default, this method returns True, so that all records are included.
        """

        return True

    def sync_related(self, instance):
        """
        Sync related objects for the given instance.
        """
        sync_classes = self.get_related_synchronizers(instance)

        for sync_class in sync_classes:
            self._relation_sync(*sync_class)

    def get_related_synchronizers(self, instance):
        """
        Return a list of related synchronizers.
        """
        return []

    @staticmethod
    def _relation_sync(synchronizer, filter_params):
        """
        Perform the sync specifically for records related to the given
        instance.
        """

        results = SyncResults()

        # Set of IDs of all records related to the parent object
        # to sync, to delete only related stale records.
        initial_ids = synchronizer.instance_ids(filter_params=filter_params)
        results = synchronizer.fetch_records(results, )

        results.deleted_count = synchronizer.prune_stale_records(
            initial_ids, results.synced_ids
        )

        return results.created_count, results.updated_count, \
            results.skipped_count, results.deleted_count



import json
import logging
import hmac
import hashlib
import base64

from braces import views
from django.views.generic import View
from django.http import HttpResponse
from django.conf import settings

from djpsa.halo.records.ticket.sync import TicketSynchronizer

logger = logging.getLogger(__name__)


class CallBackView(views.CsrfExemptMixin,
                   views.JsonRequestResponseMixin, View):
    entity_type = None
    sync_class = None
    callback_handler = None

    def post(self, request, *args, **kwargs):
        logger.debug(f"Received callback {request.path} {request.method}")

        try:
            received_token = request.headers['Token']
        except KeyError:
            logger.error('Token header missing in webhook request')
            return HttpResponse(status=401)

        computed_hash = hmac.new(
            settings.CALLBACK_SECRET.encode('utf-8'),
            request.body,
            hashlib.sha256
        ).digest()
        computed_token = base64.b64encode(computed_hash).decode()

        if not hmac.compare_digest(computed_token, received_token):
            logger.error('Invalid token received')
            return HttpResponse(status=401)

        try:
            data = json.loads(request.body.decode('utf-8'))
        except json.JSONDecodeError as e:
            logger.error('Error decoding JSON for callback: %s', e)
            return HttpResponse(status=400)

        self.handle(data)

        return HttpResponse(status=200)

    def handle(self, data):
        """
        Do the interesting stuff here, so that it can be overridden in
        a child class if needed.
        """

        if self.callback_handler:
            # Call the callback handler if it's defined
            self.callback_handler(data)
        else:
            sync = self.sync_class()

            instance, _ = sync.update_or_create_instance(
                data.get(self.entity_type))

            # Sync related records, actions, appointments, etc.
            sync.sync_related(instance)

    @classmethod
    def register_callback_handler(cls, callback_handler):
        """
        Register a callback handler for the ticket callback.
        """
        cls.callback_handler = callback_handler


class TicketCallBackView(CallBackView):

    entity_type = 'ticket'
    sync_class = TicketSynchronizer

    # This must be defined on child classes or the parent class
    # handler will be overridden for each child class.
    callback_handler = None



from djpsa.halo import models
from djpsa.halo.records import api
from djpsa.halo import sync


class StatusSynchronizer(sync.HaloSynchronizer):
    model_class = models.StatusTracker
    client_class = api.StatusAPI

    def _try_validate(self, api_instance):
        status_type = api_instance.get('type')

        # Zero is the status type for Ticket statuses, other types
        # are not for ticket. It is worth noting the API has a parameter
        # to filter by type, but it does not work.
        return status_type == 0

    def _assign_field_data(self, instance, json_data):

        instance.id = json_data.get('id')
        instance.name = json_data.get('name')
        instance.colour = json_data.get('colour')



try:
    # Redis is optional
    import redis
except ImportError:
    redis = None

import logging
from contextlib import contextmanager
from django.conf import settings

redis_client = None
logger = logging.getLogger(__name__)


def get_redis_client():
    global redis_client

    if not redis_client:
        redis_client = redis.StrictRedis(
            host=settings.REDIS['host'],
            port=settings.REDIS['port'],
            password=settings.REDIS['password'],
        )

    return redis_client


class LockNotAcquiredError(Exception):
    """The lock could not be acquired."""
    pass


@contextmanager
def redis_lock(lock_name, timeout, blocking_timeout):
    """
    Get a distributed lock using Redis.
    https://redis.readthedocs.io/en/stable/connections.html#redis.asyncio.client.Redis.lock
    """
    lock = get_redis_client().lock(
        lock_name,
        timeout=timeout,
        blocking_timeout=blocking_timeout
    )
    acquired = lock.acquire(blocking=True)
    if not acquired:
        raise LockNotAcquiredError()
    try:
        yield acquired
    finally:
        try:
            lock.release()
        except redis.exceptions.LockNotOwnedError:
            pass


def get_djpsa_settings():
    # Make some defaults
    request_settings = {
        'timeout': 30.0,
        'batch_size': 100,
        'max_attempts': 3,
        'callback_root': None,
        'callback_description': 'django-psa',
        'keep_closed_days': 1,
    }

    if hasattr(settings, 'DJPSA_CONF_CALLABLE'):
        request_settings.update(settings.DJPSA_CONF_CALLABLE())

    return request_settings



# Generated by Django 4.2.16 on 2025-02-26 16:17

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('halo', '0024_outcome_outcometracker_alter_action_outcome'),
    ]

    operations = [
        migrations.CreateModel(
            name='ChargeRate',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=255)),
            ],
        ),
        migrations.CreateModel(
            name='ChargeRateTracker',
            fields=[
            ],
            options={
                'db_table': 'halo_chargerate',
                'proxy': True,
                'indexes': [],
                'constraints': [],
            },
            bases=('halo.chargerate',),
        ),
    ]



import datetime

from django.contrib import admin

from djpsa.sync import models


@admin.register(models.SyncJob)
class SyncJobAdmin(admin.ModelAdmin):
    actions = None
    change_form_template = 'change_form.html'
    list_display = (
        'id', 'start_time', 'end_time', 'duration_or_zero', 'entity_name',
        'synchronizer_class', 'success', 'added', 'updated', 'skipped',
        'deleted', 'sync_type',
    )
    list_filter = ('sync_type', 'success', 'entity_name', 'synchronizer_class')

    def has_add_permission(self, request):
        return False

    def has_change_permission(self, request, obj=None):
        return False

    def has_delete_permission(self, request, obj=None):
        return False

    def duration_or_zero(self, obj):
        """
        Return the duration, or just the string 0 (otherwise we get 0:00:00)
        """
        duration = obj.duration()
        if duration:
            # Get rid of the microseconds part
            duration_seconds = duration - datetime.timedelta(
                microseconds=duration.microseconds
            )
            return duration_seconds if duration_seconds else '0'
    duration_or_zero.short_description = 'Duration'



# Generated by Django 4.2.17 on 2025-03-20 14:36

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('halo', '0026_merge_20250314_1651'),
    ]

    operations = [
        migrations.AlterField(
            model_name='action',
            name='charge_rate',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='halo.chargerate'),
        ),
        migrations.AlterField(
            model_name='ticket',
            name='deadline_date',
            field=models.DateTimeField(blank=True, null=True, verbose_name='Deadline'),
        ),
        migrations.AlterField(
            model_name='ticket',
            name='estimate',
            field=models.FloatField(blank=True, help_text='Effort in hours', null=True),
        ),
        migrations.AlterField(
            model_name='ticket',
            name='fix_by_date',
            field=models.DateTimeField(blank=True, null=True, verbose_name='Fix by'),
        ),
        migrations.AlterField(
            model_name='ticket',
            name='itil_request_type',
            field=models.IntegerField(choices=[(1, 'Incident'), (2, 'Change Request'), (3, 'Service Request'), (4, 'Problem'), (20, 'Request For Quote'), (21, 'Advice Other'), (22, 'Projects'), (23, 'Tasks')], default=1, verbose_name='ITIL request type'),
        ),
        migrations.AlterField(
            model_name='ticket',
            name='user',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='halo.halouser', verbose_name='Contact'),
        ),
    ]



# Generated by Django 4.2.11 on 2024-09-13 13:42

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('halo', '0007_alter_sla_track_sla_fix_by_time_and_more'),
    ]

    operations = [
        migrations.AlterField(
            model_name='site',
            name='delivery_address',
            field=models.CharField(max_length=1000),
        ),
    ]



# Generated by Django 4.2.11 on 2024-12-13 10:21

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('halo', '0017_team_teamtracker'),
    ]

    operations = [
        migrations.AddField(
            model_name='agent',
            name='teams',
            field=models.ManyToManyField(blank=True, related_name='agents', to='halo.team'),
        ),
    ]



from typing import Any, List

from django.utils import timezone
from dateutil.parser import parse

from djpsa.halo import models
from djpsa.halo.records import api
from djpsa.halo import sync
from djpsa.halo.records.action.sync import ActionSynchronizer
from djpsa.halo.records.appointment.sync import AppointmentSynchronizer
from djpsa.halo.records.agent.api import UNASSIGNED_AGENT_ID
from djpsa.halo.records.client.api import UNASSIGNED_CLIENT_ID
from djpsa.utils import get_djpsa_settings


class TicketSynchronizer(sync.ResponseKeyMixin,
                         sync.CreateMixin,
                         sync.UpdateMixin,
                         sync.DeleteMixin,
                         sync.HaloSynchronizer,
                         ):
    response_key = 'tickets'
    model_class = models.TicketTracker
    client_class = api.TicketAPI
    last_updated_field = 'lastupdatefromdate'

    related_meta = {
        'client_id': (models.Client, 'client'),
        'status_id': (models.Status, 'status'),
        'priority_id': (models.Priority, 'priority'),
        'agent_id': (models.Agent, 'agent'),
        'sla_id': (models.SLA, 'sla'),
        'user_id': (models.HaloUser, 'user'),
        'site_id': (models.Site, 'site'),
        'tickettype_id': (models.TicketType, 'type'),
        'parent_id': (models.Ticket, 'project'),
    }

    def __init__(self,
                 full: bool = False,
                 conditions: List = None,
                 *args: Any,
                 **kwargs: Any):
        super().__init__(full, conditions, *args, **kwargs)

        self.client.add_condition({
            'open_only': True,
        })

    def _assign_field_data(self, instance, json_data):
        instance.id = json_data.get('id')
        instance.summary = json_data.get('summary')
        instance.details = json_data.get('details')

        try:
            instance.last_action_date = timezone.make_aware(
                parse(json_data.get('lastactiondate')), timezone.utc)
        except ValueError:
            instance.last_action_date = parse(json_data.get('lastactiondate'))

        # Halo API has different keys for last update depending on if it's
        # a GET or POST request. This API is going to be the death of me.
        last_update = \
            json_data.get('lastupdate') or json_data.get('last_update')

        if last_update:
            try:
                instance.last_update = timezone.make_aware(
                    parse(last_update), timezone.utc)
            except ValueError:
                instance.last_update = parse(last_update)

        instance.user_email = json_data.get('useremail', instance.user_email)
        instance.reported_by = \
            json_data.get('reportedby', instance.reported_by)
        instance.end_user_status = json_data.get('enduserstatus')
        instance.category_1 = \
            json_data.get('category1') or json_data.get('category_1')
        instance.category_2 = \
            json_data.get('category2') or json_data.get('category_2')
        instance.category_3 = \
            json_data.get('category3') or json_data.get('category_3')
        instance.category_4 = \
            json_data.get('category4') or json_data.get('category_4')
        instance.inactive = json_data.get('inactive', False)
        instance.sla_response_state = json_data.get(
            'sla_response_state', instance.sla_response_state)
        instance.sla_hold_time = json_data.get('sla_hold_time',
                                               instance.sla_hold_time)
        instance.impact = json_data.get('impact')
        instance.flagged = json_data.get('flagged', False)
        instance.on_hold = json_data.get('onhold', False)
        instance.project_time_actual = \
            json_data.get('projecttimeactual', instance.project_time_actual)
        instance.project_money_actual = \
            json_data.get('projectmoneyactual', instance.project_money_actual)
        instance.cost = json_data.get('cost')
        instance.estimate = json_data.get('estimate')
        instance.estimated_days = json_data.get('estimateddays')
        instance.exclude_from_slas = json_data.get('excludefromslas', False)
        instance.reviewed = json_data.get('reviewed', False)
        instance.read = json_data.get('read', False)
        instance.use = json_data.get('use', instance.use)
        instance.email_to_list = \
            json_data.get('emailtolist', instance.email_to_list)
        instance.urgency = json_data.get('urgency', instance.urgency)
        instance.service_status_note = json_data.get('servicestatusnote')
        instance.ticket_tags = \
            json_data.get('tickettags', instance.ticket_tags)
        instance.appointment_type = json_data.get('appointment_type')
        instance.impact_level = json_data.get('impactlevel')
        instance.itil_request_type = json_data.get('itil_requesttype_id')

        date_occurred = json_data.get('dateoccurred')
        if date_occurred:
            instance.date_occurred = sync.empty_date_parser(date_occurred)

        respond_by_date = json_data.get('respondbydate')
        if respond_by_date:
            instance.respond_by_date = sync.empty_date_parser(respond_by_date)

        fix_by_date = json_data.get('fixbydate')
        if fix_by_date:
            instance.fix_by_date = sync.empty_date_parser(fix_by_date)

        date_assigned = json_data.get('dateassigned')
        if date_assigned:
            instance.date_assigned = sync.empty_date_parser(date_assigned)

        response_date = json_data.get('responsedate')
        if response_date:
            instance.response_date = sync.empty_date_parser(response_date)

        deadline_date = json_data.get('deadlinedate')
        if deadline_date:
            instance.deadline_date = sync.empty_date_parser(deadline_date)

        date_closed = json_data.get('dateclosed')
        if date_closed:
            instance.date_closed = sync.empty_date_parser(date_closed)

        start_date = json_data.get('startdate')
        if start_date:
            parsed_start_date = sync.parse_date_from_api(start_date)
            if parsed_start_date:
                instance.start_date = parsed_start_date

        target_date = json_data.get('targetdate')
        if target_date:
            parsed_target_date = sync.parse_date_from_api(target_date)
            if parsed_target_date:
                instance.target_date = parsed_target_date

        last_incoming_email_date = json_data.get('lastincomingemaildate')
        if last_incoming_email_date:
            instance.last_incoming_email_date = \
                sync.empty_date_parser(last_incoming_email_date)

        team_name = json_data.get('team')

        instance.team = models.Team.objects.filter(name=team_name).first()

        self.set_relations(instance, json_data)
        if instance.agent_id == UNASSIGNED_AGENT_ID:
            instance.agent = None
        if instance.client_id == UNASSIGNED_CLIENT_ID:
            instance.client = None

    def _post_sync_operations(self, results):
        if self.full:
            # Perform second sync for tickets that were closed
            # within the keep_closed_days setting.

            self.client.remove_condition('open_only')

            self.client.add_condition({
                'lastupdatefromdate':
                    self._get_keep_closed_cutoff().strftime(
                        '%Y-%m-%dT%H:%M:%S.%fZ')
            })
            self.client.add_condition({
                'closed_only': True,
            })

            results = self.fetch_records(results)

        return results

    def get_related_synchronizers(self, instance):
        """
        Return a list of related synchronizers.
        """
        sync_classes = []

        appointment_sync = AppointmentSynchronizer(
            full=True,
            conditions=[{
                'ticket_id': instance.id
            }]
        )

        sync_classes.append((appointment_sync, {'ticket': instance.id}))

        action_sync = ActionSynchronizer(
            full=True,
            parent_object_id=instance.id,
            conditions=[{
                'ticket_id': instance.id
            }])

        sync_classes.append((action_sync, {'ticket': instance.id}))

        return sync_classes

    def _try_validate(self, record):
        # Prevents closed tickets that were updated from re-syncing
        # every time there is an update to them.

        date_closed = record.get('dateclosed')

        if not date_closed:
            # Not closed, return true immediately.
            # Shouldn't happen, but just in case...
            return True

        true_date_closed = sync.empty_date_parser(date_closed)

        if true_date_closed:
            closed_date_cutoff = self._get_keep_closed_cutoff()

            if true_date_closed < closed_date_cutoff:
                # Closed, but older than the cutoff
                return False

        # Closed, but within the cutoff OR date_closed is not a valid date
        return True

    def _get_keep_closed_cutoff(self):
        djpsa_settings = get_djpsa_settings()
        keep_closed_days = djpsa_settings['keep_closed_days']

        return timezone.now() - timezone.timedelta(days=keep_closed_days)

    def _convert_fields(self, data):
        """
        Date fields (start_date, target_date) to Halo API format.
        """
        data = super()._convert_fields(data)

        for field_name in ['start_date', 'target_date']:
            if field_name in data and data[field_name] is not None:
                data[field_name] = sync.format_date_for_api(data[field_name])

        return data

    def _convert_fields_to_api_format(self, data):
        """Convert certain fields to the format expected by the Halo API."""
        data = super()._convert_fields_to_api_format(data)

        if 'agent_id' in data and data['agent_id'] is None:
            data['agent_id'] = UNASSIGNED_AGENT_ID

        return data



from djpsa.halo import models
from djpsa.halo.records import api
from djpsa.halo import sync


class PrioritySynchronizer(sync.HaloSynchronizer):

    model_class = models.PriorityTracker
    lookup_key = 'priorityid'
    client_class = api.PriorityAPI

    def _assign_field_data(self, instance, json_data):

        instance.id = json_data.get('priorityid')
        instance.name = json_data.get('name')
        instance.colour = json_data.get('colour')
        instance.is_hidden = json_data.get('ishidden')



import requests
import logging

from django.conf import settings
from retrying import retry
from json import JSONDecodeError

from djpsa.api import exceptions as exc
from djpsa.utils import get_djpsa_settings


RETRY_WAIT_EXPONENTIAL_MULTAPPLIER = 1000  # Initial number of milliseconds to
# wait before retrying a request.
RETRY_WAIT_EXPONENTIAL_MAX = 10000  # Maximum number of milliseconds to wait
# before retrying a request.

logger = logging.getLogger(__name__)


def retry_if_api_error(exception):
    """
    Return True if we should retry, False otherwise.

    Basically, don't retry on APIClientError, because those are the
    type of exceptions where retrying won't help (404s, 403s, etc.)
    """
    return type(exception) is exc.APIServerError


class APIClient:

    def __init__(self, conditions=None):
        self.conditions = conditions if conditions else []

        self.request_settings = get_djpsa_settings()
        if hasattr(settings, 'DJPSA_CONF_CALLABLE'):
            self.request_settings.update(
                settings.DJPSA_CONF_CALLABLE().get('request', {}))

    def add_condition(self, condition):
        self.conditions.append(condition)

    def fetch_resource(self, endpoint_url=None, params=None, should_page=False,
                       retry_counter=None,
                       *args, **kwargs):
        """
        Issue a GET request to the specified REST endpoint.

        retry_counter is a dict in the form {'count': 0} that is passed in
        to verify the number of attempts that were made.
        """

        @retry(stop_max_attempt_number=self.request_settings['max_attempts'],
               wait_exponential_multiplier=RETRY_WAIT_EXPONENTIAL_MULTAPPLIER,
               wait_exponential_max=RETRY_WAIT_EXPONENTIAL_MAX,
               retry_on_exception=retry_if_api_error)
        def _fetch_resource(endpoint_url=None, params=None, should_page=False,
                            retry_counter=None, *args, **kwargs):
            if not retry_counter:
                retry_counter = {'count': 0}
            retry_counter['count'] += 1

            return self.request('GET',
                                endpoint_url,
                                params=params,
                                *args,
                                **kwargs)

        if not retry_counter:
            retry_counter = {'count': 0}
        return _fetch_resource(endpoint_url, params=params,
                               should_page=should_page,
                               retry_counter=retry_counter,
                               *args, **kwargs)

    def request(self,
                method,
                endpoint_url=None,
                body=None,
                params=None,
                files=None,
                **kwargs):
        """
        Issue the given type of request to the specified REST endpoint.
        """
        if not endpoint_url:
            endpoint_url = self._format_endpoint()

        logger.debug(
            'Making {} request to {}, body len {}, params {}'.format(
                method, endpoint_url, len(body) if body else 0, params
            )
        )

        try:
            if body:
                kwargs['json'] = body
            if files:
                kwargs['files'] = files

            response = self._request(
                method,
                endpoint_url,
                headers=self._get_headers(),
                params=self._format_params(params),
                **kwargs
            )

        except requests.RequestException as e:
            logger.debug(
                'Request failed: {} {}: {}'.format(method, endpoint_url, e)
            )
            raise exc.APIError('{}'.format(e))

        if response.status_code == 204:  # No content
            return None
        elif 200 <= response.status_code < 300:
            # Check Content-Type to determine how to handle the response.
            # Some APIs (e.g., Halo image attachments) return binary data
            # instead of JSON.
            content_type = response.headers.get('Content-Type', '').lower()
            if 'application/json' in content_type or content_type == '':
                try:
                    return response.json()
                except JSONDecodeError as e:
                    logger.error(
                        'Request failed during decoding JSON: GET {}: {}'
                        .format(endpoint_url, e)
                    )
                    raise exc.APIError('JSONDecodeError: {}'.format(e))
            else:
                # Return the response object so caller can handle
                # appropriately
                return response
        elif response.status_code == 403:
            # TODO permissions exception from AT returns as 500, because
            #  it is terribly designed. Need to handle
            error_message = self._prepare_error_response(response)
            self._log_failed(response, error_message)
            raise exc.SecurityPermissionsException(
                error_message, response.status_code)
        elif response.status_code == 404:
            msg = 'Resource not found: {}'.format(response.url)
            logger.warning(msg)
            raise exc.RecordNotFoundError(msg)
        elif 400 <= response.status_code < 499:
            error_message = self._prepare_error_response(response)
            self._log_failed(response, error_message)
            raise exc.APIClientError(error_message)
        elif response.status_code == 500:
            error_message = self._prepare_error_response(response)
            self._log_failed(response, error_message)
            raise exc.APIServerError(error_message)
        else:
            self._log_failed(response, response.content)
            raise exc.APIError(response)

    def _log_failed(self, response, message):
        logger.error(f'Failed request: HTTP {response.status_code} '
                     f'for {response.url}; response {message}')

    def _prepare_error_response(self, response):
        return response.json().content

    def _get_request_kwargs(self):
        raise NotImplementedError('Subclasses must implement this method.')

    def _get_request_decorator(self):
        raise NotImplementedError('Subclasses must implement this method.')

    def _request(self,
                 method,
                 endpoint_url,
                 headers=None,
                 params=None,
                 **kwargs):
        raise NotImplementedError('Subclasses must implement this method.')

    def _format_endpoint(self):
        raise NotImplementedError('Subclasses must implement this method.')

    def _format_params(self, params=None):
        raise NotImplementedError('Subclasses must implement this method.')

    def get_page(self, page=None, batch_size=None, params=None):
        raise NotImplementedError('Subclasses must implement this method.')

    def _get_headers(self):
        return {}



# Generated by Django 4.2.11 on 2024-08-15 10:33

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('halo', '0001_initial'),
    ]

    operations = [
        migrations.RenameField(
            model_name='agent',
            old_name='isdisabled',
            new_name='is_disabled',
        ),
        migrations.RenameField(
            model_name='priority',
            old_name='ishidden',
            new_name='is_hidden',
        ),
        migrations.RenameField(
            model_name='ticket',
            old_name='lastactiondate',
            new_name='last_action_date',
        ),
        migrations.RenameField(
            model_name='ticket',
            old_name='targetdate',
            new_name='target_date',
        ),
        migrations.RenameField(
            model_name='ticket',
            old_name='targettime',
            new_name='target_time',
        ),
    ]



# This file is used to import all the api classes for the records module
# from a single location.

from djpsa.halo.records.ticket.api import TicketAPI
from djpsa.halo.records.priority.api import PriorityAPI
from djpsa.halo.records.status.api import StatusAPI
from djpsa.halo.records.client.api import ClientAPI
from djpsa.halo.records.agent.api import AgentAPI
from djpsa.halo.records.site.api import SiteAPI
from djpsa.halo.records.sla.api import SLAAPI
from djpsa.halo.records.appointment.api import AppointmentAPI
from djpsa.halo.records.halouser.api import UserAPI
from djpsa.halo.records.tickettype.api import TicketTypeAPI
from djpsa.halo.records.action.api import ActionAPI
from djpsa.halo.records.team.api import TeamAPI
from djpsa.halo.records.budgettype.api import BudgetTypeAPI
from djpsa.halo.records.budgetdata.api import BudgetDataAPI
from djpsa.halo.records.outcome.api import OutcomeAPI
from djpsa.halo.records.chargerate.api import ChargeRateAPI
from djpsa.halo.records.timesheetevent.api import TimesheetEventAPI



# Generated by Django 4.2.11 on 2024-09-13 13:36

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('halo', '0005_sla_site_halouser_ticket_sla'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='sla',
            name='hours_are_local_time',
        ),
    ]



# Generated by Django 4.2.11 on 2024-08-14 14:34

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Agent',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=255)),
                ('isdisabled', models.BooleanField(default=False)),
                ('email', models.EmailField(blank=True, max_length=254, null=True)),
                ('initials', models.CharField(blank=True, max_length=10, null=True)),
                ('firstname', models.CharField(blank=True, max_length=255, null=True)),
                ('surname', models.CharField(blank=True, max_length=255, null=True)),
                ('colour', models.CharField(blank=True, max_length=255, null=True)),
            ],
        ),
        migrations.CreateModel(
            name='Client',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('inactive', models.BooleanField(default=False)),
                ('name', models.CharField(max_length=255)),
            ],
        ),
        migrations.CreateModel(
            name='Priority',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(blank=True, max_length=255, null=True)),
                ('ishidden', models.BooleanField(default=False)),
                ('colour', models.CharField(blank=True, max_length=255, null=True)),
            ],
        ),
        migrations.CreateModel(
            name='Status',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=255)),
                ('colour', models.CharField(blank=True, max_length=255, null=True)),
            ],
        ),
        migrations.CreateModel(
            name='Ticket',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('summary', models.CharField(max_length=255)),
                ('details', models.TextField(blank=True, max_length=8000, null=True)),
                ('lastactiondate', models.DateField(blank=True, null=True)),
                ('last_update', models.DateTimeField(blank=True, null=True)),
                ('targetdate', models.DateField(blank=True, null=True)),
                ('targettime', models.TimeField(blank=True, null=True)),
                ('agent', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='halo.agent')),
                ('client', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='halo.client')),
                ('priority', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='halo.priority')),
                ('status', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='halo.status')),
            ],
        ),
    ]



# Generated by Django 4.2.17 on 2025-03-14 16:51

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('halo', '0024_ticket_date_closed'),
        ('halo', '0025_chargerate_chargeratetracker'),
    ]

    operations = [
    ]



from typing import Any, List
from djpsa.halo import models
from djpsa.halo.records import api
from djpsa.halo import sync


class SiteSynchronizer(sync.ResponseKeyMixin, sync.HaloSynchronizer):

    model_class = models.SiteTracker
    response_key = 'sites'
    client_class = api.SiteAPI

    related_meta = {
        'client_id': (models.Client, 'client'),
        'sla_id': (models.SLA, 'sla'),
    }

    def __init__(self,
                 full: bool = False,
                 conditions: List = None,
                 *args: Any,
                 **kwargs: Any):
        super().__init__(full, conditions, *args, **kwargs)

        self.client.add_condition({
            'includeaddress': True,
        })

    def _assign_field_data(self, instance, json_data):

        instance.id = json_data.get('id')
        instance.name = json_data.get('name')
        # api has client_id as decimal, convert to int. It's never
        # actually a decimal. The API docs has it as an int, but for some
        # reason it's being returned as a decimal on Site records.
        if json_data.get('client_id'):
            json_data['client_id'] = int(json_data.get('client_id'))

        # This is the worst thing I have ever seen.
        instance.delivery_address = json_data.get('delivery_address_line1')
        instance.delivery_address += \
            " {}".format(json_data.get('delivery_address_line2'))
        instance.delivery_address += \
            " {}".format(json_data.get('delivery_address_line3'))
        instance.delivery_address += \
            " {}".format(json_data.get('delivery_address_line4'))
        instance.delivery_address += \
            " {}".format(json_data.get('delivery_address_line5'))

        instance.colour = json_data.get('colour')
        instance.active = json_data.get('active', False)
        instance.phone_number = json_data.get('phone_number')
        instance.use = json_data.get('use')

        self.set_relations(instance, json_data)



# Generated by Django 4.2.11 on 2024-09-13 13:38

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('halo', '0006_remove_sla_hours_are_local_time'),
    ]

    operations = [
        migrations.AlterField(
            model_name='sla',
            name='track_sla_fix_by_time',
            field=models.BooleanField(default=False),
        ),
        migrations.AlterField(
            model_name='sla',
            name='track_sla_response_time',
            field=models.BooleanField(default=False),
        ),
    ]



from django.db import models
from model_utils import FieldTracker


class Appointment(models.Model):
    subject = models.CharField(max_length=1000, blank=True, null=True)
    start_date = models.DateTimeField(blank=True, null=True)
    end_date = models.DateTimeField(blank=True, null=True)
    appointment_type = models.CharField(max_length=255, blank=True, null=True)
    is_private = models.BooleanField(default=False)
    # Appointments have two types: appointment and task. Task
    # in this context does NOT mean a project task.
    is_task = models.BooleanField(default=False)
    complete_status = models.IntegerField(blank=True, null=True)
    colour = models.CharField(max_length=50, null=True, blank=True)
    # Not using URLField because I don't trust API data.
    online_meeting_url = \
        models.CharField(max_length=2000, blank=True, null=True)
    ticket = models.ForeignKey(
        'Ticket', on_delete=models.CASCADE, blank=True, null=True)
    site = models.ForeignKey(
        'Site', on_delete=models.SET_NULL, blank=True, null=True)
    agent = models.ForeignKey(
        'Agent', on_delete=models.CASCADE, blank=True, null=True)
    user = models.ForeignKey(
        'HaloUser', on_delete=models.CASCADE, blank=True, null=True)
    client = models.ForeignKey(
        'Client', on_delete=models.CASCADE, blank=True, null=True)

    API_FIELDS = {
        'id': 'id',
        'subject': 'subject',
        'start_date': 'start_date',
        'end_date': 'end_date',
        'appointment_type': 'appointment_type_name',
        'is_private': 'is_private',
        'is_task': 'is_task',
        'complete_status': 'complete_status',
        'colour': 'colour',
        'online_meeting_url': 'online_meeting_url',
        'ticket': 'ticket_id',
        'site': 'site_id',
        'agent': 'agent_id',
        'user': 'user_id',
        'client': 'client_id',
    }

    class Meta:
        verbose_name_plural = "Appointments"

    def __str__(self):
        return f"Appointment {self.id} - {self.subject}"


class AppointmentTracker(Appointment):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'halo_appointment'



# Generated by Django 4.2.11 on 2024-08-16 11:50

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('halo', '0003_alter_ticket_agent_alter_ticket_client_and_more'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='agent',
            options={'verbose_name_plural': 'Agents'},
        ),
        migrations.AlterModelOptions(
            name='client',
            options={'verbose_name_plural': 'Clients'},
        ),
        migrations.AlterModelOptions(
            name='priority',
            options={'verbose_name_plural': 'Priorities'},
        ),
        migrations.AlterModelOptions(
            name='status',
            options={'verbose_name_plural': 'Statuses'},
        ),
        migrations.AlterModelOptions(
            name='ticket',
            options={'verbose_name_plural': 'Tickets'},
        ),
    ]



from django.apps import AppConfig
from django.conf import settings


class HaloConfig(AppConfig):
    default_auto_field = 'django.db.models.BigAutoField'
    name = 'djpsa.halo'
    label = 'halo'

    def ready(self):
        from djpsa.halo import provider
        settings.PROVIDER = provider



# This file is used to import all the synchronizers for the records module
# from a single location.

from django.utils.translation import gettext_lazy as _

from djpsa.halo.records.ticket.sync import TicketSynchronizer
from djpsa.halo.records.priority.sync import PrioritySynchronizer
from djpsa.halo.records.status.sync import StatusSynchronizer
from djpsa.halo.records.client.sync import ClientSynchronizer
from djpsa.halo.records.agent.sync import AgentSynchronizer
from djpsa.halo.records.site.sync import SiteSynchronizer
from djpsa.halo.records.sla.sync import SLASynchronizer
from djpsa.halo.records.appointment.sync import AppointmentSynchronizer
from djpsa.halo.records.halouser.sync import HaloUserSynchronizer
from djpsa.halo.records.tickettype.sync import TicketTypeSynchronizer
from djpsa.halo.records.action.sync import ActionSynchronizer
from djpsa.halo.records.team.sync import TeamSynchronizer
from djpsa.halo.records.budgettype.sync import BudgetTypeSynchronizer
from djpsa.halo.records.budgetdata.sync import BudgetDataSynchronizer
from djpsa.halo.records.outcome.sync import OutcomeSynchronizer
from djpsa.halo.records.chargerate.sync import ChargeRateSynchronizer
from djpsa.halo.records.timesheetevent.sync import TimeSheetEventSynchronizer

from djpsa.sync.grades import SyncGrades


class HaloSyncGrades(SyncGrades):
    def partial_grades(self):
        return [
            TicketSynchronizer,
        ]

    def operational_grades(self):
        return [
            HaloUserSynchronizer,
            AgentSynchronizer,
            ClientSynchronizer,
            TicketSynchronizer,
            AppointmentSynchronizer,
            ActionSynchronizer,
        ]

    def configuration_grades(self):
        """
        Return a list of synchronizers for resources that change infrequently-
        such as on a weekly or monthly basis. For example, ticket types, statuses,
        priorities, etc.
        """
        return [
            SLASynchronizer,
            StatusSynchronizer,
            SiteSynchronizer,
            PrioritySynchronizer,
            TicketTypeSynchronizer,
            ChargeRateSynchronizer,
            OutcomeSynchronizer,
            TeamSynchronizer,
            BudgetTypeSynchronizer,
            BudgetDataSynchronizer,
        ]


sync_command_list = [
        ('status', (StatusSynchronizer, _('Status'))),
        ('priority', (PrioritySynchronizer, _('Priority'))),
        ('client', (ClientSynchronizer, _('Client'))),
        ('sla', (SLASynchronizer, _('SLA'))),
        ('site', (SiteSynchronizer, _('Site'))),
        ('user', (HaloUserSynchronizer, _('User'))),
        ('agent', (AgentSynchronizer, _('Agent'))),
        ('chargerate', (ChargeRateSynchronizer, _('ChargeRate'))),
        ('outcome', (OutcomeSynchronizer, _('Outcome'))),
        ('ticket_type', (TicketTypeSynchronizer, _('TicketType'))),
        ('ticket', (TicketSynchronizer, _('Ticket'))),
        ('appointment', (AppointmentSynchronizer, _('Appointment'))),
        ('action', (ActionSynchronizer, _('Action'))),
        ('team', (TeamSynchronizer, _('Team'))),
        ('budget_type', (BudgetTypeSynchronizer, _('BudgetType'))),
        ('budget_data', (BudgetDataSynchronizer, _('BudgetData'))),
    ]



from djpsa.halo.api import HaloAPIClient


class ChargeRateAPI(HaloAPIClient):
    # Including parameter in the endpoint because we never
    # want this API class to access a different lookup id.
    endpoint = 'Lookup?lookupid=17'



# Generated by Django 4.2.11 on 2024-10-21 12:18

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('halo', '0014_alter_action_project_alter_action_ticket'),
    ]

    operations = [
        migrations.AlterField(
            model_name='ticket',
            name='details',
            field=models.TextField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name='ticket',
            name='email_to_list',
            field=models.TextField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name='ticket',
            name='service_status_note',
            field=models.TextField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name='ticket',
            name='ticket_tags',
            field=models.TextField(blank=True, null=True),
        ),
    ]



# Generated by Django 4.2.11 on 2024-09-19 15:00

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('halo', '0008_alter_site_delivery_address'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='halouser',
            options={'verbose_name_plural': 'Halo Users'},
        ),
        migrations.AlterModelOptions(
            name='site',
            options={'verbose_name_plural': 'Sites'},
        ),
        migrations.AlterModelOptions(
            name='sla',
            options={'verbose_name_plural': 'SLAs'},
        ),
        migrations.AddField(
            model_name='ticket',
            name='category_1',
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
        migrations.AddField(
            model_name='ticket',
            name='category_2',
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
        migrations.AddField(
            model_name='ticket',
            name='category_3',
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
        migrations.AddField(
            model_name='ticket',
            name='category_4',
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
        migrations.AddField(
            model_name='ticket',
            name='cost',
            field=models.FloatField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='ticket',
            name='date_assigned',
            field=models.DateTimeField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='ticket',
            name='date_occurred',
            field=models.DateTimeField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='ticket',
            name='deadline_date',
            field=models.DateTimeField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='ticket',
            name='email_to_list',
            field=models.TextField(blank=True, max_length=2000, null=True),
        ),
        migrations.AddField(
            model_name='ticket',
            name='end_user_status',
            field=models.IntegerField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='ticket',
            name='estimate',
            field=models.FloatField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='ticket',
            name='estimated_days',
            field=models.FloatField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='ticket',
            name='exclude_from_sla',
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name='ticket',
            name='fix_by_date',
            field=models.DateTimeField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='ticket',
            name='flagged',
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name='ticket',
            name='impact',
            field=models.IntegerField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='ticket',
            name='inactive',
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name='ticket',
            name='last_incoming_email',
            field=models.DateTimeField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='ticket',
            name='on_hold',
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name='ticket',
            name='project_money_actual',
            field=models.FloatField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='ticket',
            name='project_time_actual',
            field=models.FloatField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='ticket',
            name='read',
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name='ticket',
            name='reported_by',
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
        migrations.AddField(
            model_name='ticket',
            name='respond_by_date',
            field=models.DateTimeField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='ticket',
            name='response_date',
            field=models.DateTimeField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='ticket',
            name='reviewed',
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name='ticket',
            name='service_status_note',
            field=models.TextField(blank=True, max_length=2000, null=True),
        ),
        migrations.AddField(
            model_name='ticket',
            name='site',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='halo.site'),
        ),
        migrations.AddField(
            model_name='ticket',
            name='sla_hold_time',
            field=models.FloatField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='ticket',
            name='sla_response_state',
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
        migrations.AddField(
            model_name='ticket',
            name='start_date',
            field=models.DateField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='ticket',
            name='start_time',
            field=models.TimeField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='ticket',
            name='team',
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
        migrations.AddField(
            model_name='ticket',
            name='ticket_tags',
            field=models.TextField(blank=True, max_length=2000, null=True),
        ),
        migrations.AddField(
            model_name='ticket',
            name='urgency',
            field=models.IntegerField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='ticket',
            name='use',
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
        migrations.AddField(
            model_name='ticket',
            name='user',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='halo.halouser'),
        ),
        migrations.AddField(
            model_name='ticket',
            name='user_email',
            field=models.EmailField(blank=True, max_length=254, null=True),
        ),
        migrations.AlterField(
            model_name='halouser',
            name='first_name',
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name='halouser',
            name='initials',
            field=models.CharField(blank=True, max_length=10, null=True),
        ),
        migrations.AlterField(
            model_name='halouser',
            name='login',
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name='halouser',
            name='name',
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name='halouser',
            name='surname',
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name='halouser',
            name='use',
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name='site',
            name='delivery_address',
            field=models.CharField(blank=True, max_length=1000, null=True),
        ),
        migrations.AlterField(
            model_name='site',
            name='use',
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
        migrations.CreateModel(
            name='Project',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('summary', models.CharField(blank=True, max_length=1000, null=True)),
                ('details', models.TextField(blank=True, max_length=8000, null=True)),
                ('appointment_type', models.CharField(blank=True, max_length=255, null=True)),
                ('closed_in_integration_system', models.BooleanField(default=False)),
                ('cost', models.FloatField(blank=True, null=True)),
                ('date_assigned', models.DateTimeField(blank=True, null=True)),
                ('date_occurred', models.DateTimeField(blank=True, null=True)),
                ('deadline_date', models.DateTimeField(blank=True, null=True)),
                ('estimate', models.FloatField(blank=True, null=True)),
                ('estimate_days', models.FloatField(blank=True, null=True)),
                ('exclude_from_sla', models.BooleanField(default=False)),
                ('fix_by_date', models.DateTimeField(blank=True, null=True)),
                ('flagged', models.BooleanField(default=False)),
                ('impact', models.IntegerField(blank=True, null=True)),
                ('impact_level', models.IntegerField(blank=True, null=True)),
                ('inactive', models.BooleanField(default=False)),
                ('last_action_date', models.DateTimeField(blank=True, null=True)),
                ('last_incoming_email', models.DateTimeField(blank=True, null=True)),
                ('on_hold', models.BooleanField(default=False)),
                ('project_time_actual', models.FloatField(blank=True, null=True)),
                ('respond_by_date', models.DateTimeField(blank=True, null=True)),
                ('reviewed', models.BooleanField(default=False)),
                ('source', models.IntegerField(blank=True, null=True)),
                ('start_date', models.DateField(blank=True, null=True)),
                ('start_time', models.TimeField(blank=True, null=True)),
                ('target_date', models.DateField(blank=True, null=True)),
                ('target_time', models.TimeField(blank=True, null=True)),
                ('team', models.CharField(blank=True, max_length=255, null=True)),
                ('urgency', models.IntegerField(blank=True, null=True)),
                ('use', models.CharField(blank=True, max_length=255, null=True)),
                ('agent', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='halo.agent')),
                ('client', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='halo.client')),
                ('priority', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='halo.priority')),
                ('site', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='halo.site')),
                ('status', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='halo.status')),
                ('user', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='halo.halouser')),
            ],
            options={
                'verbose_name_plural': 'Appointments',
            },
        ),
        migrations.CreateModel(
            name='Appointment',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('subject', models.CharField(blank=True, max_length=1000, null=True)),
                ('start_date', models.DateTimeField(blank=True, null=True)),
                ('end_date', models.DateTimeField(blank=True, null=True)),
                ('appointment_type', models.CharField(blank=True, max_length=255, null=True)),
                ('is_private', models.BooleanField(default=False)),
                ('is_task', models.BooleanField(default=False)),
                ('complete_status', models.IntegerField(blank=True, null=True)),
                ('colour', models.CharField(blank=True, max_length=50, null=True)),
                ('online_meeting_url', models.CharField(blank=True, max_length=2000, null=True)),
                ('agent', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='halo.agent')),
                ('client', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='halo.client')),
                ('site', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='halo.site')),
                ('ticket', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='halo.ticket')),
                ('user', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='halo.halouser')),
            ],
            options={
                'verbose_name_plural': 'Appointments',
            },
        ),
        migrations.AddField(
            model_name='ticket',
            name='parent',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='halo.project'),
        ),
    ]



# Generated by Django 4.2.11 on 2024-12-13 09:55

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('halo', '0016_alter_appointment_site_alter_client_main_site_and_more'),
    ]

    operations = [
        migrations.CreateModel(
            name='Team',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=255)),
                ('ticket_count', models.IntegerField(default=0)),
            ],
            options={
                'verbose_name_plural': 'Teams',
            },
        ),
        migrations.CreateModel(
            name='TeamTracker',
            fields=[
            ],
            options={
                'db_table': 'halo_team',
                'proxy': True,
                'indexes': [],
                'constraints': [],
            },
            bases=('halo.team',),
        ),
    ]



class SyncGrades:
    """
    Define grades of synchronizers.

    The result of operational+configuration+slow grades should be all the
    synchronizers.
    """
    def partial_grades(self):
        """
        Return a list of synchronizers for resources that support partial syncs
        and that are useful to keep up-to-date at a high frequency.
        """
        return []

    def operational_grades(self):
        """
        Return a list of synchronizers for resources that change throughout a
        typical day. For example, tickets, service calls, notes, etc.

        Exclude resources that can potentially take a very long time to sync.
        """
        return []

    def configuration_grades(self):
        """
        Return a list of synchronizers for resources that change infrequently-
        such as on a weekly or monthly basis. For example, ticket types,
        statuses, priorities, etc.
        """
        return []

    def slow_grades(self):
        """
        Return a list of synchronizers for resources that can potentially take
        a very long time to sync. For example, notes, time entries, etc.
        """
        return []



# Generated by Django 4.2.11 on 2024-09-13 13:33

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('halo', '0004_alter_agent_options_alter_client_options_and_more'),
    ]

    operations = [
        migrations.CreateModel(
            name='SLA',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=255)),
                ('hours_are_local_time', models.BooleanField(default=False)),
                ('hours_are_techs_local_time', models.BooleanField(default=False)),
                ('response_reset', models.BooleanField(default=False)),
                ('response_reset_approval', models.BooleanField(default=False)),
                ('track_sla_fix_by_time', models.BooleanField(default=True)),
                ('track_sla_response_time', models.BooleanField(default=True)),
                ('workday_id', models.IntegerField(blank=True, null=True)),
                ('auto_release_limit', models.IntegerField(blank=True, null=True)),
                ('auto_release_option', models.BooleanField(default=False)),
                ('status_after_first_warning', models.IntegerField(blank=True, null=True)),
                ('status_after_second_warning', models.IntegerField(blank=True, null=True)),
                ('status_after_auto_release', models.IntegerField(blank=True, null=True)),
            ],
        ),
        migrations.CreateModel(
            name='Site',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=255)),
                ('colour', models.CharField(blank=True, max_length=50, null=True)),
                ('active', models.BooleanField(default=True)),
                ('phone_number', models.CharField(blank=True, max_length=250, null=True)),
                ('use', models.CharField(max_length=255)),
                ('delivery_address', models.BooleanField(max_length=1000)),
                ('client', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='halo.client')),
                ('sla', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='halo.sla')),
            ],
        ),
        migrations.CreateModel(
            name='HaloUser',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=255)),
                ('first_name', models.CharField(max_length=255)),
                ('initials', models.CharField(max_length=10)),
                ('surname', models.CharField(max_length=255)),
                ('email', models.EmailField(blank=True, max_length=250, null=True)),
                ('colour', models.CharField(blank=True, max_length=50, null=True)),
                ('active', models.BooleanField(default=True)),
                ('login', models.CharField(max_length=255)),
                ('use', models.CharField(max_length=255)),
                ('never_send_emails', models.BooleanField(default=False)),
                ('phone_number', models.CharField(blank=True, max_length=250, null=True)),
                ('mobile_number', models.CharField(blank=True, max_length=250, null=True)),
                ('mobile_number_2', models.CharField(blank=True, max_length=250, null=True)),
                ('home_number', models.CharField(blank=True, max_length=250, null=True)),
                ('tel_pref', models.IntegerField(blank=True, null=True)),
                ('is_service_account', models.BooleanField(default=False)),
                ('is_important_contact', models.BooleanField(default=False)),
                ('is_important_contact_2', models.BooleanField(default=False)),
                ('agent', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='halo.agent')),
                ('client', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='halo.client')),
                ('site', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='halo.site')),
            ],
        ),
        migrations.AddField(
            model_name='ticket',
            name='sla',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='halo.sla'),
        ),
    ]



from django.core.management.base import BaseCommand, CommandError
from django.utils.translation import gettext_lazy as _

from djpsa.api import exceptions as exc

OPTION_NAME = 'sync_object'


class BaseSyncCommand(BaseCommand):
    help = 'Synchronize'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def add_arguments(self, parser):
        parser.add_argument(OPTION_NAME, nargs='?', type=str)
        parser.add_argument('--full',
                            action='store_true',
                            dest='full',
                            default=False)

    def sync_by_class(self, sync_class, obj_name, full_option=False):
        synchronizer = sync_class(full=full_option)

        created_count, updated_count, skipped_count, deleted_count = \
            synchronizer.sync()

        msg = _('{} Sync Summary - Created: {}, Updated: {}, Skipped: {}')
        fmt_msg = msg.format(obj_name, created_count, updated_count,
                             skipped_count)

        if full_option:
            msg = _('{} Sync Summary - Created: {}, Updated: {}, '
                    'Skipped: {}, Deleted: {}')
            fmt_msg = msg.format(obj_name, created_count, updated_count,
                                 skipped_count, deleted_count)

        self.stdout.write(fmt_msg)

    def handle(self, *args, **options):
        sync_classes = []
        object_arg = options[OPTION_NAME]
        full_option = options.get('full', False)

        if object_arg:
            object_arg = object_arg
            sync_tuple = self.synchronizer_map.get(object_arg)

            if sync_tuple:
                sync_classes.append(sync_tuple)
            else:
                msg = _('Invalid object {}, '
                        'choose one of the following: \n{}')
                options_txt = ', '.join(self.synchronizer_map.keys())
                msg = msg.format(sync_tuple, options_txt)
                raise CommandError(msg)
        else:
            sync_classes = self.synchronizer_map.values()

        failed_classes = 0
        error_messages = ''

        for sync_class, obj_name in sync_classes:
            try:
                self.sync_by_class(sync_class, obj_name,
                                   full_option=full_option)
            except exc.SecurityPermissionsException as e:
                msg = 'Failed to sync {}: {}'.format(obj_name, e)
                self.stderr.write(msg)
                error_messages += '{}\n'.format(msg)

            except exc.APIError as e:
                msg = 'Failed to sync {}: {}'.format(obj_name, e)
                self.stderr.write(msg)
                error_messages += '{}\n'.format(msg)
                failed_classes += 1

        if failed_classes > 0:
            msg = '{} class{} failed to sync.\n'.format(
                failed_classes,
                '' if failed_classes == 1 else 'es',
            )
            msg += 'Errors:\n'
            msg += error_messages
            raise CommandError(msg)



import logging

logger = logging.getLogger(__name__)


class CallbacksHandler:
    api_client = None
    field_names = None
    ident_field = None
    base_callback_data = None

    def __init__(self):
        self.client = self.api_client()

    @property
    def needed_callbacks(self):
        raise NotImplementedError

    def _build_get_conditions(self):
        raise NotImplementedError

    def _build_post_data(self, callback):
        raise NotImplementedError

    def _post_register_processing(self, registered_callback):
        pass

    def _register_callback(self, callback):
        logger.info('Registering {} callback'.format(
            callback[self.ident_field]))
        registered_callback = self.client.create(
            self._build_post_data(callback))

        # Do anything else that needs to be done after registering a callback,
        #  depending on the PSA
        self._post_register_processing(registered_callback)

    def _delete_callback(self, callback):
        logger.info('Deleting callback {}'.format(callback['id']))
        self.client.delete(callback['id'])

    def _calculate_missing_unneeded_callbacks(self,
                                              needed_callbacks,
                                              current_callbacks):
        """
        Given the list of needed callbacks and current callbacks, figure out
        which callbacks need to be registered and which ones need to be
        removed.

        Returns a tuple of (need-to-add, need-to-remove) callbacks.
        """
        # For each current callback that matches a needed callback, delete it
        # from both needed and current.
        current_indexes_delete = set()
        needed_indexes_delete = set()

        # Add the index of the element to delete to a list, because it's
        # an error to delete elements out of a list while iterating through.
        for i_cur, current in enumerate(current_callbacks):
            for i_need, needed in enumerate(needed_callbacks):
                all_matched = True
                for field in self.field_names:
                    if current[field] != needed[field]:
                        # If any field doesn't match, don't delete the cb
                        all_matched = False
                if all_matched:
                    current_indexes_delete.add(i_cur)
                    needed_indexes_delete.add(i_need)
        for i_cur in reversed(sorted(current_indexes_delete)):
            # Delete from the end
            current_callbacks.pop(i_cur)
        for i_need in reversed(sorted(needed_indexes_delete)):
            # Delete from the end
            needed_callbacks.pop(i_need)

        return needed_callbacks, current_callbacks

    def get_callbacks(self):
        callbacks = self.client.get(**self._build_get_conditions())

        # Don't scan through any callbacks if they aren't ours
        cleaned_callbacks = self._clean_callbacks(callbacks)

        return cleaned_callbacks

    def get_needed_callbacks(self):
        """
        Return a list of callbacks
        """
        result = self.needed_callbacks
        for cb in result:
            cb['url'] = f"{self.settings['callback_root']}" \
                        f"{cb['url']}"

        return result

    def _clean_callbacks(self, callbacks):
        return callbacks

    def ensure_registered(self):
        """
        Do the minimum changes to ensure our callbacks are registered
        exactly once.
        """
        needed_callbacks = self.get_needed_callbacks()
        current_callbacks = self.get_callbacks()

        callbacks_to_add, callbacks_to_remove = \
            self._calculate_missing_unneeded_callbacks(
                needed_callbacks, current_callbacks
            )

        for callback in callbacks_to_add:
            self._register_callback(callback)
        for callback in callbacks_to_remove:
            self._delete_callback(callback)

    def ensure_deleted(self):
        """Do the needful to ensure our callbacks are gone."""
        callbacks = self.get_callbacks()

        for callback in callbacks:
            self._delete_callback(callback)



from django.db import models
from model_utils import FieldTracker


class Client(models.Model):
    name = models.CharField(max_length=255)
    inactive = models.BooleanField(default=False)
    colour = models.CharField(max_length=255, blank=True, null=True)
    main_site = models.ForeignKey('Site',
                                  blank=True,
                                  null=True,
                                  on_delete=models.SET_NULL,
                                  related_name='client_primary'
                                  )
    phone_number = models.CharField(blank=True, null=True, max_length=250)

    class Meta:
        verbose_name_plural = "Clients"

    def __str__(self):
        return str(self.name)


class ClientTracker(Client):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'halo_client'



import logging

from django.conf import settings

from djpsa.halo.api import WebhookAPIClient, NotificationAPIClient
from djpsa.utils import get_djpsa_settings
from djpsa.sync.callbacks import CallbacksHandler

logger = logging.getLogger(__name__)

event_data = {
    'faults': [{
        "webhook_id": None,
        "type": "-2",
        "delivery_method": "6",
        "name": "Ticket updated",
        "agent_id": "0",
        "slack_id": "0",
        "teams_id": "0",
        "filter_type": "2",
        "restriction_type": "0",
        "popupinnotificationpane": False,
        "sendpushnotificationbrowser": False,
        "sendpushnotification": False,
        "customisecolour": False,
        "eventno": "61",
        "interval": "0",
        "useworkinghours": "0",
        "emailtemplate_id": "32",
        "conditions": [
            {
                "fieldname": "symptom",
                "value_type": "string",
                "tablename": "faults",
                "type": 1,
                "change_context": 0,
                "value_string": '""',
                "value_lookup": None,
                "value_display": "Incident",
            }
        ],
    }, {
        "webhook_id": None,
        "type": "-2",
        "delivery_method": "6",
        "name": "",
        "agent_id": "0",
        "slack_id": "0",
        "teams_id": "0",
        "email_list": "",
        "filter_type": "2",
        "restriction_type": "0",
        "sendpushnotification": False,
        "customisecolour": False,
        "eventno": "3",
        "interval": "0",
        "useworkinghours": "0",
        "conditions": [],
        "emailtemplate_id": "32"
    }]
}


class HaloCallbacksHandler(CallbacksHandler):
    api_client = WebhookAPIClient
    field_names = ['name', 'note', 'url']
    ident_field = 'name'
    base_callback_data = {
        'active': True,
        'algorithm': 0,
        'authentication_type': 0,
        'certificate_id': 0,
        'content_type': 'application/json',
        'inbound_authentication_type': 0,
        'method': 0,
        'major_version_number': 0,
        'minor_version_number': 1,
        'patch_version_number': 0,
        'runbook_start_type': 0,
        'type': 0,
        'systemuse': '',
        'version_number': '0.0.1'
    }

    def __init__(self):
        self.settings = get_djpsa_settings()
        self.event_api = NotificationAPIClient()
        super().__init__()

    @property
    def needed_callbacks(self):
        # NOTE: Halo doesn't give you anything to identify callbacks as your
        #  callbacks, so the 'callback_description' should be used to filter
        #  out callbacks that you shouldn't delete.
        callback_description = self.settings['callback_description']

        return [
            {
                'name': f'{callback_description} Ticket Callback',
                'note': 'faults',
                'url': "ticket/",
            },
        ]

    def _build_get_conditions(self):
        return {}

    def _build_post_data(self, callback):

        auth_data = {
            'authentication_header': "Token",
            'basic_password': settings.CALLBACK_SECRET,
            'authentication_type': "3",
        }

        callback = {
            **callback,
            **auth_data,
        }

        callback_post_data = {**self.base_callback_data, **callback}
        return callback_post_data

    def _clean_callbacks(self, callbacks):

        # Create a new list of elements from the callbacks list
        # that where the name field contains the string field
        # defined in 'callback_description'.

        names_to_check = [cb['name'] for cb in self.get_needed_callbacks()]

        return [
            cb for cb in callbacks
            if 'name' in cb
               and any(name in cb['name'] for name in names_to_check)
        ]

    def _post_register_processing(self, registered_callback):
        # Add events to the callback
        events = event_data.get(registered_callback['note'])

        if events:
            for event in events:
                event['webhook_id'] = registered_callback['id']
            self.event_api.create(data=events)
        else:
            logger.exception(
                f"No event template found for {registered_callback['note']}")



from typing import Any, List

from djpsa.halo import models
from djpsa.halo.records import api
from djpsa.halo import sync


class AgentSynchronizer(sync.ResponseKeyMixin,
                        sync.HaloSynchronizer):

    model_class = models.AgentTracker
    client_class = api.AgentAPI
    response_key = "results"

    def __init__(
            self,
            full: bool = False,
            conditions: List = None,
            *args: Any,
            **kwargs: Any):
        super().__init__(full, conditions, *args, **kwargs)
        self.client.add_condition({
            'includeenabled': True,
            'includedisabled': True,
            'includeunassigned': False,
        })

    def _assign_field_data(self, instance, json_data):

        instance.id = json_data.get('id')
        instance.name = json_data.get('name')
        instance.is_disabled = json_data.get('isdisabled')
        instance.email = json_data.get('email')
        instance.initials = json_data.get('initials')
        instance.firstname = json_data.get('firstname')
        instance.surname = json_data.get('surname')
        instance.colour = json_data.get('colour')

        teams = json_data.get('teams')
        if teams and instance.tracker.previous('id') is not None:
            # Teams can only be saved if the agent instance
            # already exists in the database. If it doesn't exist
            # yet, we'll let it be saved now and then next time this runs
            # the teams will be set.

            # Get the list of team IDs
            remote_team_ids = [team['team_id'] for team in teams]

            # Get the teams from the db
            local_teams = models.Team.objects.filter(id__in=remote_team_ids)

            # Set the teams for the agent instance, this handles adding
            # and removing teams from the agent instance.
            instance.teams.set(local_teams)



class APICondition:
    """A single API condition"""

    _formatter = None

    VALID_OPERATORS = {"and", "or", "==", "!=", ">=", "<=", ">", "<"}

    def __init__(self, *args, op=None, field=None, value=None):
        self._items = []
        self.op = op
        self.field = field
        self.value = value

        if self.op not in self.VALID_OPERATORS:
            raise ValueError(
                f"Invalid operator: {self.op}. "
                f"Must be one of {self.VALID_OPERATORS}")

        # If grouping (and/or), ensure the args are instances of APICondition
        if self.op in ("and", "or"):
            for condition in args:
                if not isinstance(condition, self.__class__):
                    raise TypeError(
                        f"Grouped conditions must also "
                        f"be instances of {self.__class__.__name__}"
                    )
                self._items.append(condition)

    def __repr__(self):
        if len(self._items):
            return f"GroupedCondition(op={self.op}, conditions={self._items})"
        return \
            f"Condition(op={self.op}, field={self.field}, value={self.value})"

    def format_condition(self):
        """Formats condition using the class-level formatter"""
        if self._formatter:
            return self._formatter.format(self)
        raise NotImplementedError("Formatter not provided for API condition.")

    def is_grouped(self):
        """Checks if the condition is grouped (AND/OR)"""
        return len(self._items) > 0


class APIConditionList:
    """Holds a list of API conditions and formats them"""

    _formatter = None

    def __init__(self):
        self._list = []

    def __getitem__(self, i):
        return self._list[i]

    def __setitem__(self, i, value):
        if not isinstance(value, APICondition):
            raise TypeError("Conditions must be instances of APICondition.")
        self._list[i] = value

    def __repr__(self):
        return f"APIConditionList({self._list})"

    def __len__(self):
        return len(self._list)

    def __delitem__(self, i):
        del self._list[i]

    def __iter__(self):
        return iter(self._list)

    def append(self, condition):
        """Add a new condition to the list"""
        if not isinstance(condition, APICondition):
            raise TypeError("Condition must be an instance of APICondition.")
        self._list.append(condition)

    def build_query(self, method="get", **kwargs):
        """Builds the query by formatting all conditions in the list"""
        queries = []
        for condition in self._list:
            queries.append(condition.format_condition())
        return self._formatter.build_query(queries, method=method, **kwargs)



from django.db import models
from model_utils import FieldTracker


class TicketType(models.Model):
    name = models.CharField(max_length=255)
    agents_can_select = models.BooleanField(default=False)
    end_users_can_select = models.BooleanField(default=False)
    allow_attachments = models.BooleanField(default=False)
    can_create = models.BooleanField(default=False)
    copy_attachments_to_child = models.BooleanField(default=False)
    copy_attachments_to_related = models.BooleanField(default=False)
    use = models.CharField(max_length=255, blank=True, null=True)

    def __str__(self):
        return str(self.name)


class TicketTypeTracker(TicketType):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'halo_tickettype'



# Generated by Django 4.2.16 on 2025-06-18 10:03

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('halo', '0028_merge_20250321_1529'),
    ]

    operations = [
        migrations.AlterField(
            model_name='ticket',
            name='appointment_type',
            field=models.IntegerField(blank=True, null=True),
        ),
    ]



from djpsa.halo import models
from djpsa.halo.records import api
from djpsa.halo import sync


class TicketTypeSynchronizer(sync.HaloSynchronizer):
    model_class = models.TicketTypeTracker
    client_class = api.TicketTypeAPI

    def _assign_field_data(self, instance, json_data):
        instance.id = json_data.get('id')
        instance.name = json_data.get('name')
        instance.description = json_data.get('description')
        instance.active = json_data.get('active', False)
        instance.use = json_data.get('use')
        instance.colour = json_data.get('colour')



# Generated by Django 4.2.16 on 2025-01-23 15:17

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('halo', '0021_alter_budgetdata_options'),
    ]

    operations = [
        migrations.AlterField(
            model_name='budgetdata',
            name='ticket',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='halo.ticket'),
        ),
    ]



# Generated by Django 4.2.11 on 2025-01-17 13:49

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('halo', '0019_alter_ticket_team'),
    ]

    operations = [
        migrations.CreateModel(
            name='BudgetType',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(blank=True, max_length=255, null=True)),
                ('default_rate', models.DecimalField(blank=True, decimal_places=2, max_digits=10, null=True)),
            ],
            options={
                'verbose_name_plural': 'Budget Types',
            },
        ),
        migrations.AddField(
            model_name='ticket',
            name='itil_request_type',
            field=models.IntegerField(choices=[(1, 'Incident'), (2, 'Change Request'), (3, 'Service Request'), (4, 'Problem'), (20, 'Request For Quote'), (21, 'Advice Other'), (22, 'Projects'), (23, 'Tasks')], default=1),
        ),
        migrations.CreateModel(
            name='BudgetData',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('hours', models.DecimalField(blank=True, decimal_places=2, max_digits=10, null=True)),
                ('rate', models.DecimalField(blank=True, decimal_places=2, max_digits=10, null=True)),
                ('money', models.DecimalField(blank=True, decimal_places=2, max_digits=10, null=True)),
                ('estimated_hours', models.DecimalField(blank=True, decimal_places=2, max_digits=10, null=True)),
                ('estimated_money', models.DecimalField(blank=True, decimal_places=2, max_digits=10, null=True)),
                ('actual_hours', models.DecimalField(blank=True, decimal_places=2, max_digits=10, null=True)),
                ('actual_money', models.DecimalField(blank=True, decimal_places=2, max_digits=10, null=True)),
                ('scheduled_hours', models.DecimalField(blank=True, decimal_places=2, max_digits=10, null=True)),
                ('scheduled_value', models.DecimalField(blank=True, decimal_places=2, max_digits=10, null=True)),
                ('toschedule_hours', models.DecimalField(blank=True, decimal_places=2, max_digits=10, null=True)),
                ('toschedule_value', models.DecimalField(blank=True, decimal_places=2, max_digits=10, null=True)),
                ('remaining_hours', models.DecimalField(blank=True, decimal_places=2, max_digits=10, null=True)),
                ('remaining_value', models.DecimalField(blank=True, decimal_places=2, max_digits=10, null=True)),
                ('budget_type', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='halo.budgettype')),
                ('ticket', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='halo.ticket')),
            ],
        ),
        migrations.CreateModel(
            name='BudgetDataTracker',
            fields=[
            ],
            options={
                'db_table': 'halo_budgetdata',
                'proxy': True,
                'indexes': [],
                'constraints': [],
            },
            bases=('halo.budgetdata',),
        ),
        migrations.CreateModel(
            name='BudgetTypeTracker',
            fields=[
            ],
            options={
                'db_table': 'halo_budgettype',
                'proxy': True,
                'indexes': [],
                'constraints': [],
            },
            bases=('halo.budgettype',),
        ),
    ]



from django.db import models


class SyncJob(models.Model):
    start_time = models.DateTimeField(null=False)
    end_time = models.DateTimeField(blank=True, null=True)
    entity_name = models.CharField(max_length=100)
    synchronizer_class = models.CharField(max_length=100, blank=True,
                                          null=True)
    added = models.PositiveIntegerField(null=True)
    updated = models.PositiveIntegerField(null=True)
    deleted = models.PositiveIntegerField(null=True)
    skipped = models.PositiveIntegerField(null=True)
    success = models.BooleanField(null=True)
    message = models.TextField(blank=True, null=True)
    sync_type = models.CharField(max_length=32, default='full')

    def duration(self):
        if self.start_time and self.end_time:
            return self.end_time - self.start_time
